'''''''''''''''''''''''''''''''''''''''''''''
' Jos: AD-Doku
' VB-Script JoseExec.vbs zum Ausfhren der Dokumentation
' Version 3.12 / 08.04.2019
' Tool zur Dokumentation von Active Directory
' Listet die wichtigsten AD-Objekte grafisch
' in einer HTML-Datei auf.
'
' Von Nils Kaczenski, Ansgar Wiechers und Nils Weinhold
'
' Keine Gewhr! Nutzung auf eigene Gefahr!
'
' Letzte nderungen: siehe changelog.txt
'''''''''''''''''''''''''''''''''''''''''''''

Option Explicit

'Region Constants
Const AppName    = "Jos AD-Dokumentation"
Const AppVersion = "3.12"
Const Authors    = "Nils Kaczenski, Ansgar Wiechers und Nils Weinhold"
Const Homepage   = "http://www.faq-o-matic.net/jose/"

Const ForReading         = 1
Const TristateUseDefault = -2

Const KEY   = 0
Const VALUE = 1

' See <http://msdn.microsoft.com/en-us/library/ms680832.aspx>
Const ADS_UF_ACCOUNTDISABLE = &h00000002
Const ADS_UF_LOCKOUT        = &h00000010

' See <http://msdn.microsoft.com/en-us/library/cc232505.aspx>
Const GPO_IGNORE  = &h00000001
Const GPO_ENFORCE = &h00000002

Const NTDSDSA_OPT_IS_GC = 1

' See <http://msdn.microsoft.com/en-us/library/cc245476(PROT.13).aspx>
Const DOMAIN_PASSWORD_COMPLEX = &h00000001
Const DOMAIN_PASSWORD_NO_CLEAR_CHANGE = &h00000004
Const DOMAIN_PASSWORD_STORE_CLEARTEXT = &h00000010

Const ResourceDir      = "Dateien/"
Const ReportStyleSheet = "report.css"
Const ReportDir        = "Reports"
Const DefinitionDir    = "Definitions"
Const ReportSuffix     = ".htm"
Const IndentionStep    = 20 'px ("Tab" width for indention)

'EndRegion

'Region Global variables
Dim fso : Set fso = CreateObject("Scripting.FileSystemObject")

Dim rootDSE
Dim adoCmd
Dim hasADSIErrorMessages
Dim systemBias
Dim maxOULevel : maxOULevel = 1

Dim showPermissionHint
Dim PermissionReason : PermissionReason = Array()

Dim arrUserAttr
Dim arrComputerAttr
Dim arrGroupAttr
Dim arrContactAttr
Dim arrPrinterAttr
'EndRegion

'Region Default configuration values
Dim cfg : Set cfg = CreateObject("Scripting.Dictionary")
	cfg.Add "title", Null
	cfg.Add "reportFilename", Null
	cfg.Add "definitionFilename", Null
	cfg.Add "generator", AppName & " v" & AppVersion
	cfg.Add "includeDomainDetails", False
	cfg.Add "includeTrusts", False
	cfg.Add "includeUPNSuffixes", False
	cfg.Add "includeOUs", False
	cfg.Add "includeUsers", False
	cfg.Add "includeComputers", False
	cfg.Add "includeGroups", False
	cfg.Add "includeContacts", False
	cfg.Add "includePrinters", False
	cfg.Add "includeGPOs", False
	cfg.Add "includeCounts", False
	cfg.Add "includeDates", False
	cfg.Add "includePSOs", False
	cfg.Add "showDescriptions", False
	cfg.Add "showFriendlyNames", False
	cfg.Add "showNaturalNames", False
	cfg.Add "showCaption", False
	cfg.Add "showParent", False
	cfg.Add "objectFilter", ""
	cfg.Add "debugEnabled", False
'EndRegion

'Region AD Constants
' See <http://msdn.microsoft.com/en-us/library/ms680325(VS.85).aspx>
Dim trustFlags : Set trustFlags = CreateObject("Scripting.Dictionary")
	trustFlags.Add &h00000001, "TRUST_ATTRIBUTE_NON_TRANSITIVE" ' Disallow transitivity
	trustFlags.Add &h00000002, "TRUST_ATTRIBUTE_UPLEVEL_ONLY" ' Trust link only valid For uplevel client
	trustFlags.Add &h00000004, "TRUST_ATTRIBUTE_QUARANTINED_DOMAIN" ' Used To quarantine domains
	trustFlags.Add &h00000008, "TRUST_ATTRIBUTE_FOREST_TRANSITIVE" ' This link may contain forest trust information
	trustFlags.Add &h00000010, "TRUST_ATTRIBUTE_CROSS_ORGANIZATION" ' This trust Is To a domain/forest which Is Not part of this enterprise
	trustFlags.Add &h00000020, "TRUST_ATTRIBUTE_WITHIN_FOREST" ' Trust Is internal To this forest
	trustFlags.Add &h00000040, "TRUST_ATTRIBUTE_TREAT_AS_EXTERNAL" ' Trust Is To be treated As external For trust boundary purposes
	trustFlags.Add &h00000080, "TRUST_ATTRIBUTE_TRUST_USES_RC4_ENCRYPTION" ' MIT trust With RC4
	' Trust attributes 0x00000040 through 0x00200000 are reserved For future use
	' Trust attributes 0x00400000 through 0x00800000 were used previously (up To W2K) And should Not be re-used
	' Trust attributes 0x01000000 through 0x80000000 are reserved For user	
	trustFlags.Add &hFF03FFFF, "TRUST_ATTRIBUTES_VALID" ' bit mask to determine if trust attributes are valid
	trustFlags.Add &hFF000000, "TRUST_ATTRIBUTES_USER" ' bit mask to determine if user attributes are set

' See <http://msdn.microsoft.com/en-us/library/ms680832.aspx>
' corrections inspired by Philipp Fckeler
Dim uacFlags : Set uacFlags = CreateObject("Scripting.Dictionary")
	uacFlags.Add &h00000001, "ADS_UF_SCRIPT"
	uacFlags.Add &h00000002, "ADS_UF_ACCOUNTDISABLE"
	uacFlags.Add &h00000008, "ADS_UF_HOMEDIR_REQUIRED"
	uacFlags.Add &h00000010, "ADS_UF_LOCKOUT"
	uacFlags.Add &h00000020, "ADS_UF_PASSWD_NOTREQD"
	uacFlags.Add &h00000040, "ADS_UF_PASSWD_CANT_CHANGE"
	uacFlags.Add &h00000080, "ADS_UF_ENCRYPTED_TEXT_PASSWORD_ALLOWED"
	uacFlags.Add &h00000100, "ADS_UF_TEMP_DUPLICATE_ACCOUNT"
	uacFlags.Add &h00000200, "ADS_UF_NORMAL_ACCOUNT"
	uacFlags.Add &h00000800, "ADS_UF_INTERDOMAIN_TRUST_ACCOUNT"
	uacFlags.Add &h00001000, "ADS_UF_WORKSTATION_TRUST_ACCOUNT"
	uacFlags.Add &h00002000, "ADS_UF_SERVER_TRUST_ACCOUNT"
	uacFlags.Add &h00010000, "ADS_UF_DONT_EXPIRE_PASSWD"
	uacFlags.Add &h00020000, "ADS_UF_MNS_LOGON_ACCOUNT"
	uacFlags.Add &h00040000, "ADS_UF_SMARTCARD_REQUIRED"
	uacFlags.Add &h00080000, "ADS_UF_TRUSTED_FOR_DELEGATION"
	uacFlags.Add &h00100000, "ADS_UF_NOT_DELEGATED"
	uacFlags.Add &h00200000, "ADS_UF_USE_DES_KEY_ONLY"
	uacFlags.Add &h00400000, "ADS_UF_DONT_REQUIRE_PREAUTH"
	uacFlags.Add &h00800000, "ADS_UF_PASSWORD_EXPIRED"
	uacFlags.Add &h01000000, "ADS_UF_TRUSTED_TO_AUTHENTICATE_FOR_DELEGATION"
	uacFlags.Add &h02000000, "ADS_ UF_NO_AUTH_DATA_REQUIRED"
	uacFlags.Add &h04000000, "ADS_ UF_PARTIAL_SECRETS_ACCOUNT"
'EndRegion

'Region Script Defaults
' Dictionary for translating attribute names into (user-)friendly names.
Dim naturalName : Set naturalName = CreateObject("Scripting.Dictionary")
	naturalName.Add "altrecipient", "Mail-Weiterleitung"
	naturalName.Add "altrecipientcopy", "Mail-Weiterleitung (in Kopie)"
	naturalName.Add "msexchgenericforwardingaddress", "SMTP-Weiterleitung"
	naturalName.Add "msexchgenericforwardingaddresscopy", "SMTP-Weiterleitung (in Kopie)"
	naturalName.Add "canonicalname", "OU-Pfad (canonicalName)"
	naturalName.Add "company", "Firma"
	naturalName.Add "dcroles", "Computer-Rolle"
	naturalName.Add "department", "Abteilung"
	naturalName.Add "distinguishedname", "LDAP-Name (DN)"
	naturalName.Add "mailnickname", "Mail-Alias"
	naturalName.Add "givenname", "Vorname"
	naturalName.Add "grouptype", "Gruppentyp"
	naturalName.Add "homedirectory", "persnliches Verzeichnis"
	naturalName.Add "homedrive", "persnliches Laufwerk"
	naturalName.Add "location", "Standort"
	naturalName.Add "logonhours", "Anmeldezeiten"
	naturalName.Add "mail", "E-Mail-Adresse"
	naturalName.Add "member", "Mitglieder"
	naturalName.Add "memberof", "Mitglied von"
	naturalName.Add "msds-psoapplied", "Gltige PSOs"
	naturalName.Add "msds-resultantpso", "Effektives PSO"
	naturalName.Add "msexchhomeservername", "Exchange-Homeserver"
	naturalName.Add "mstshomedirectory", "Terminaldienste-Home (msTS*)"   ' Server 2008 only - unused! (see: <http://msdn.microsoft.com/en-us/library/cc220562.aspx>)
	naturalName.Add "mstsprofilepath", "Terminaldienste-Profil (msTS*)"   ' Server 2008 only - unused! (see: <http://msdn.microsoft.com/en-us/library/cc220579.aspx>)
	naturalName.Add "objectsid", "SID"
	naturalName.Add "operatingsystem", "Betriebssystem"
	naturalName.Add "operatingsystemservicepack", "Service-Pack"
	naturalName.Add "operatingsystemversion", "Betriebssystemversion"
	naturalName.Add "physicaldeliveryofficename", "Bro"
	naturalName.Add "primarygroupid", "Primre Gruppe (ID)"
	naturalName.Add "primarygrouptoken", "Token der Primren Gruppe"
	naturalName.Add "printsharename", "Freigabename"
	naturalName.Add "profilepath", "servergespeichertes Profil"
	naturalName.Add "proxyaddresses", "Exchange-E-Mail-Adressen"
	naturalName.Add "pwdlastset", "Kennwort gendert"
	naturalName.Add "samaccountname", "Anmeldename"
	naturalName.Add "scriptpath", "Anmeldeskript"
	naturalName.Add "servername", "Druckserver"
	naturalName.Add "showinaddressbook", "Anzeige in Adressbuch"
	naturalName.Add "sidhistory", "SID-Verlauf"
	naturalName.Add "sn", "Nachname"
	naturalName.Add "streetaddress", "Adresse"
	naturalName.Add "telephonenumber", "Telefon"
	naturalName.Add "terminalserviceshomedirectory", "Terminaldienste-Home (userParameters)" 
	naturalName.Add "terminalservicesprofilepath", "Terminaldienste-Profil (userParameters)" 
	naturalName.Add "useraccountcontrol", "Kontenattribute"
	naturalName.Add "userprincipalname", "UPN-Anmeldename"
	naturalName.Add "whenchanged", "Objekt gendert am"
	naturalName.Add "whencreated", "Objekt erzeugt am"

' Symbol tags
Dim symbols : Set symbols = CreateObject("Scripting.Dictionary")
	symbols.Add "Domain", "<img src=""" & ResourceDir & "domain.gif"">&nbsp;"
	symbols.Add "organizationalUnit", "<img src=""" & ResourceDir & "ou.gif"">&nbsp;"
	symbols.Add "container", "<img src=""" & ResourceDir & "ordner.gif"">&nbsp;"
	symbols.Add "builtinDomain", "<img src=""" & ResourceDir & "ordner.gif"">&nbsp;"
	symbols.Add "User", "<img src=""" & ResourceDir & "user.gif"">&nbsp;"
	symbols.Add "Computer", "<img src=""" & ResourceDir & "computer.gif"">&nbsp;"
	symbols.Add "Group", "<img src=""" & ResourceDir & "group.gif"">&nbsp;"
	symbols.Add "Contact", "<img src=""" & ResourceDir & "contact.gif"">&nbsp;"
	symbols.Add "Printer", "<img src=""" & ResourceDir & "printer.gif"">&nbsp;"
	symbols.Add "GPO", "<img src=""" & ResourceDir & "gpo.gif"">&nbsp;"
	symbols.Add "PSO", "<img src=""" & ResourceDir & "pso.gif"">&nbsp;"
	symbols.Add "disabled", "<img src=""" & ResourceDir & "disable.gif"">&nbsp;"
	symbols.Add "locked", "<img src=""" & ResourceDir & "locked.gif"">&nbsp;"
	symbols.Add "Question", "<img src=""" & ResourceDir & "question.gif"">&nbsp;"
	symbols.Add "Exclamation", "<img src=""" & ResourceDir & "exclamation.gif"">&nbsp;"
	symbols.Add "blocked", "<img src=""" & ResourceDir & "blocked.gif"">&nbsp;"
	symbols.Add "Clock", "<img src=""" & ResourceDir & "uhr.gif"">&nbsp;"
	symbols.Add "Jos", "<img src=""" & ResourceDir & "jose-klein.gif"" alt=""Jos Active-Directory-Dokumentation"" title=""Jos Active-Directory-Dokumentation"">"
	symbols.Add "FOM", "<img src=""" & ResourceDir & "fom-klein.gif"" alt=""http://www.faq-o-matic.net"" title=""http://www.faq-o-matic.net"">"

Dim stats : Set stats = CreateObject("Scripting.Dictionary")
	stats.Add "user", 0
	stats.Add "computer", 0
	stats.Add "printQueue", 0
	stats.Add "group", 0
	stats.Add "contact", 0
	stats.Add "organizationalUnit", 0
	' we don't use container in v3.1 any more
	' stats.Add "container", 0
'EndRegion

Main WScript.Arguments


Sub Main(args)
	Dim domain
	
	If InStr(LCase(WScript.FullName), "cscript") = 0 Then
		MsgBox "Dieses Script kann nicht mit WScript.exe gestartet werden." & vbNewLine _
			& "Bitte starten Sie es auf der Kommandozeile mit folgendem Kommando:" & vbNewLine _
			& vbNewLine & "cscript //u """ & WScript.ScriptFullName & """ /?" _
			, vbInformation, WScript.ScriptName & " - Mit CScript.exe starten"
		WScript.Quit 1
	End If
	
	If WScript.Version < "5.6" Then
		WScript.Echo "Dieses Skript bentigt mindestens den Windows Script Host 5.6." & vbNewLine _
			& "Die Version Ihres WSH ist: " & WScript.Version & vbNewLine _
			& "Bitte installieren Sie die aktuelle Version, die Sie bei Microsoft" & vbNewLine _
			& "herunterladen knnen."
		WScript.Quit 1
	End If

	If args.count = 0 Then
		PrintSyntax
	End If

	On Error Resume Next
	Set rootDSE = GetObject("LDAP://rootDSE")
	domain = rootDSE.Get("DefaultNamingContext")
	If TrapError("No domain available") Then WScript.Quit 1
	On Error Goto 0

	systemBias = GetSystemBias
	hasADSIErrorMessages = CheckForADSIErrorMessages
	showPermissionHint = False

	' Open AD connection only once.
	Dim adoConn : Set adoConn = CreateObject("ADODB.Connection")
	adoConn.Provider = "ADsDSOObject"
	adoConn.Open "Active Directory Provider"

	Set adoCmd = CreateObject("ADODB.Command")
	adoCmd.ActiveConnection = adoConn
	adoCmd.Properties("Page Size") = 100
	adoCmd.Properties("Timeout") = 30
	adoCmd.Properties("Cache Results") = False

	ParseCmd(args)
	ReadDefinitions cfg("definitionFilename")

	CreateReport domain

	adoConn.Close
	Set adoConn = Nothing
	Set adoCmd = Nothing
End Sub

' ==============================================================================
' Configuration functions
' ==============================================================================

' ------------------------------------------------------------------------------
' Parse the commandline arguments.
' ------------------------------------------------------------------------------
Sub ParseCmd(args)
	If args.Named.Exists("?") Then
		PrintSyntax
	Else
		With args.Named
			' Definition filename (optional)
			' If filename is "pipe" or is omitted, definitions will be read
			' from STDIN.
			If .Exists("d") And .Item("d") <> "" Then
				If LCase(.Item("d")) <> "pipe" Then cfg("definitionFilename") = .Item("d")
			End If
			' note: should there be no title or report file name they will be generated in CreateReport!
			' Report filename (optional)
			If .Exists("r") And .Item("r") <> "" Then 
				cfg("reportFilename") = .Item("r")
			End If
			' Report title (optional)
			If .Exists("t") And .Item("t") <> "" Then cfg("title") = .Item("t")
			' Object filter (optional)
			If .Exists("f") And .Item("f") <> "" Then cfg("objectFilter") = .Item("f")
			If .Exists("debug") Then cfg("debugEnabled") = True
		End With
	End If

	PrintDebug "Definitionsdatei: " & cfg("definitionFilename")
	PrintDebug "Reportdatei: " & cfg("reportFilename")
	PrintDebug "Titel: " & cfg("title")
	PrintDebug "Filter: " & cfg("objectFilter")
	PrintDebug "System Time Zone Bias: " & GetSystemBias

End Sub

' ------------------------------------------------------------------------------
' Read definitions of what to include in the report.
' ------------------------------------------------------------------------------
Sub ReadDefinitions(filename)
	Dim defFile, objFile, line, section
	Dim userAttribs, computerAttribs, groupAttribs, contactAttribs, printerAttribs

	Dim userAttributes     : Set userAttributes = CreateAttributeSet()
	Dim computerAttributes : Set computerAttributes = CreateAttributeSet()
	Dim groupAttributes    : Set groupAttributes = CreateAttributeSet()
	Dim contactAttributes  : Set contactAttributes = CreateAttributeSet()
	Dim printerAttributes  : Set printerAttributes = CreateAttributeSet()

	' Read definitions from STDIN unless a definition file is specified.
	If IsNull(filename) Then
		Set objFile = WScript.StdIn
		PrintDebug "Read definitions from STDIN ..."
	Else
		defFile = fso.BuildPath(fso.BuildPath(fso.GetParentFolderName(WScript.ScriptFullName), DefinitionDir), filename)
		If Not fso.FileExists(defFile) Then
			WScript.StdErr.WriteLine "Definition file '" & defFile & "' does not exist!"
			WScript.Quit 1
		End If
		Set objFile = fso.OpenTextFile(defFile, ForReading)
		PrintDebug "Read definitions from '" & defFile & "' ..."
	End If

	section = ""
	Do Until objFile.AtEndOfStream
		line = Trim(Replace(objFile.ReadLine, vbTab, " "))

		If Left(line, 1) = "[" Then
			PrintDebug line
			section = line
			Select Case LCase(section)
				Case "[user]" cfg("includeUsers") = True
				Case "[computer]" cfg("includeComputers") = True
				Case "[group]" cfg("includeGroups") = True
				Case "[contact]" cfg("includeContacts") = True
				Case "[printer]" cfg("includePrinters") = True
			End Select
		ElseIf line = "" Or Left(line, 1) = ";" Then
			' continue (line is either blank or a comment)
		Else
			PrintDebug line
			Select Case LCase(section)
				Case "[meta]"
					Dim def : def = EvalDefinition(line)
					' Use report title and filename from definition file only if they
					' haven't been set already via commandline.
					Select Case LCase(def(KEY))
						' note: should there be no title or report file name they will be generated below!
						Case "reportname" If IsNull(cfg("title")) Then cfg("title") = def(VALUE)
						Case "reportfilename" If IsNull(cfg("reportFilename")) Then cfg("reportFilename") = def(VALUE)
						Case "objectfilter" cfg("objectFilter") = def(VALUE)
						Case "showsymbols" cfg("showCaption") = def(VALUE)
						Case "shownaturalname" cfg("showNaturalNames") = def(VALUE)
						Case "showfriendlyname" cfg("showFriendlyNames") = def(VALUE)
						' Case "showdescriptions" cfg("showDescriptions") = def(VALUE)
						Case Else WScript.StdErr.WriteLine "Unknown definition in " & section & ": " & line
					End Select
				Case "[common]"
					Select Case LCase(line)
						Case "fsmo" cfg("includeDomainDetails") = True
						Case "trust" cfg("includeTrusts") = True
						Case "upnsuffixes" cfg("includeUPNSuffixes") = True
						Case "objectreport" cfg("includeOUs") = True
						Case "gpo" cfg("includeGPOs") = True
						Case "number" cfg("includeCounts") = True
						Case "modify" cfg("includeDates") = True
						Case "description" cfg("showDescriptions") = True
						Case "parent" cfg("showParent") = True
						Case "pso" cfg("includePSOs") = True
						Case "permission" cfg("showAdvancedPermissionHint") = True
						Case Else WScript.StdErr.WriteLine "Unknown definition in " & section & ": " & line
					End Select
				Case "[user]"
					If Not userAttributes.Exists(line) Then userAttributes.Add line, True
					' If properties for MSTS are included, add versions both for
					' Windows Server 2003 and 2008 as they use different attributes
					If LCase(line) = "terminalserviceshomedirectory" Then
						userAttributes.Add "mstshomedirectory", True
					End If
					If LCase(line) = "terminalservicesprofilepath" Then
						userAttributes.Add "mstsprofilepath", True
					End If
				Case "[computer]"
					If Not computerAttributes.Exists(line) Then computerAttributes.Add line, True
				Case "[group]"
					If Not groupAttributes.Exists(line) Then groupAttributes.Add line, True
				Case "[contact]"
					If Not contactAttributes.Exists(line) Then contactAttributes.Add line, True
				Case "[printer]"
					If Not printerAttributes.Exists(line) Then printerAttributes.Add line, True
				Case Else WScript.StdErr.WriteLine "Unknown section name: " & section
			End Select
		End If
	Loop

	if IsNull(filename) Then objFile.Close ' don't close STDIN
	Set objFile = Nothing
	
	' add Password Setting Object attributes for users if needed
	If cfg("includePSOs") Then 
		userAttributes.Add "msDS-PSOApplied", True
		userAttributes.Add "msDS-ResultantPSO", True
	End If

	' Set report title and filename to default values if they haven't been
	' set via commandline or definition file.
	If IsEmpty(cfg("title")) Then cfg("title") = DefaultTitle & " " & CStr(Now)
	If IsEmpty(cfg("reportFilename")) Then cfg("reportFilename") = Replace(Replace(cfg("title"), ":", "-"), "/", "-") & ReportSuffix

	arrUserAttr = userAttributes.Keys
	arrComputerAttr = computerAttributes.Keys
	arrGroupAttr = groupAttributes.Keys
	arrContactAttr = contactAttributes.Keys
	arrPrinterAttr = printerAttributes.Keys

	Set userAttributes = Nothing
	Set computerAttributes = Nothing
	Set groupAttributes = Nothing
	Set contactAttributes = Nothing
	Set printerAttributes = Nothing

	PrintDebug "Current Definitions:"
	Dim setting
	For Each setting In cfg.Keys
		PrintDebug setting & ":" & vbTab & cfg(setting)
	Next

	PrintDebug "User attributes: " & Join(arrUserAttr, ",")
	PrintDebug "Computer attributes: " & Join(arrComputerAttr, ",")
	PrintDebug "Group attributes: " & Join(arrGroupAttr, ",")
	PrintDebug "Contact attributes: " & Join(arrContactAttr, ",")
	PrintDebug "Printer attributes: " & Join(arrPrinterAttr, ",")
End Sub

' ==============================================================================
' Report generation functions
' ==============================================================================

' ------------------------------------------------------------------------------
' Create the report for the given domain.
' ------------------------------------------------------------------------------
Sub CreateReport(domain)
	Dim reportFile, objReport, objDomain, folder, policies, domainInfo

	PrintDebug "Create report " & cfg("reportFilename") & " ..."

	folder = fso.BuildPath(fso.GetParentFolderName(WScript.ScriptFullName), ReportDir)
	If Not fso.FolderExists(folder) Then
		fso.CreateFolder folder
		PrintDebug "Created folder: " & folder
	End If

	On Error Resume Next
	reportFile = fso.BuildPath(folder, cfg("reportFilename"))
	Set objReport = fso.CreateTextFile(reportFile,,vbTrue)
	If TrapError("Error creating file '" & reportFile & "'") Then WScript.Quit 1
	On Error Goto 0

	' write HTML header
	objReport.WriteLine "<!DOCTYPE HTML PUBLIC ""-//W3C//DTD HTML 4.01 Transitional//EN" & vbNewLine _
		& vbTab & """http://www.w3.org/TR/html4/loose.dtd"">" & vbNewLine _
		& "<HTML>" & vbNewLine & "<HEAD>" & vbNewLine _
		& "<TITLE>" & cfg("title") & "</TITLE>" & vbNewLine _
		& "<META NAME=""generator"" CONTENT=""" & cfg("generator") _
		& " (von " & Authors & ")"">" & vbNewLine _
		& "<LINK REL=""stylesheet"" TYPE=""text/css"" HREF=""" & ResourceDir _
		& ReportStyleSheet & """>" & vbNewLine & "</HEAD>" & vbNewLine & "<BODY>" _
		& vbNewLine & "<H1 id=""title"">" & symbols("Jos") & "&nbsp;" & cfg("title") & "</H1>"
	If cfg("objectFilter") <> "" Then
		objReport.WriteLine "<P class=""filter"">Ausgabe beschrnkt auf Elemente des Containers: <em>" & cfg("objectFilter") & "</em></P>"
	End If

	If cfg("includeGPOs") Then 
		Set policies = GetGPOs(domain)
	Else
		Set policies = Nothing
	End If

	' we need to always read the domain data as we may need the details later
	Set domainInfo = GetDomainInfo(domain)

	WriteDomainInfo objReport, domainInfo
	If cfg("includeTrusts") Then WriteTrusts objReport, GetTrusts(domain)
	if cfg("includePSOs") then WritePSOs objReport, GetPSOs(domain)
	If cfg("includeOUs") Then
		WriteOUInfo objReport, GetOUs("builtinDomain", domain, 0), policies, domainInfo("lockoutDuration")
		WriteOUInfo objReport, GetOUs("container", domain, 0), policies, domainInfo("lockoutDuration")
		WriteOUInfo objReport, GetOUs("organizationalUnit", domain, 0), policies, domainInfo("lockoutDuration")
	End If
	If cfg("includeGPOs") Then WriteGPOInfo objReport, policies
	If cfg("showCaption") Then WriteCaption objReport
	WriteAdv objReport

	Set policies = Nothing
	Set domainInfo = Nothing

	' write HTML footer
	objReport.WriteLine "</BODY>" & vbNewLine & "</HTML>"

	PrintDebug "Finished."

	objReport.Close
	Set objReport = Nothing
End Sub

' ------------------------------------------------------------------------------
' Write information of the given domain object to the report.
' ------------------------------------------------------------------------------
Sub WriteDomainInfo(report, domainInfo)
	Dim i
	
	report.Write "<strong>" & symbols("Domain") & domainInfo("fqdn")
	WriteDescription report, domainInfo("description")
	report.WriteLine "</strong>"

	If Not cfg("includeDomainDetails") Then Exit Sub

	report.WriteLine "<TABLE style=""margin-left:" & IndentionStep & "px;""><TR>"

	If cfg("includeDates") Then
		report.WriteLine "<TR><TD>Objekt erzeugt am</TD><TD class=""value"">" & domainInfo("whenCreated") & "</TD></TR>"
		report.WriteLine "<TR><TD>Objekt gendert am</TD><TD class=""value"">" & domainInfo("whenChanged") & "</TD></TR>"
	End If

	report.WriteLine "<TR><TD>Zuletzt gesichert</TD><TD class=""value"">" & domainInfo("lastBackup")& "</TD></TR>"
	report.WriteLine "<TR><TD>Abgefragter DC</TD><TD class=""value"">" & domainInfo("DCtoQuery") & "</TD></TR>"
	report.WriteLine "<TR><TD>Root-Domne (Forest)</TD><TD class=""value"">" & domainInfo("rootDomain") & "</TD></TR>"

	' Write FSMO role owners.
	report.WriteLine "<TR><TD>Schema-Master (Forest)</TD><TD class=""value"">" & domainInfo("fsmoSchemaMaster") & "</TD></TR>"
	report.WriteLine "<TR><TD>DNS-Master (Forest)</TD><TD class=""value"">" & domainInfo("fsmoDnsMaster") & "</TD></TR>"
	report.WriteLine "<TR><TD>PDC-Emulator (Domne)</TD><TD class=""value"">" & domainInfo("fsmoPdcEmulator") & "</TD></TR>"
	report.WriteLine "<TR><TD>RID-Master (Domne)</TD><TD class=""value"">" & domainInfo("fsmoRidMaster") & "</TD></TR>"
	report.WriteLine "<TR><TD>Infrastruktur-Master (Domne)</TD><TD class=""value"">" & domainInfo("fsmoInfraMaster") & "</TD></TR>"

	' Write operation modes.
	report.WriteLine "<TR><TD>Forest-Betriebsmodus</TD><TD class=""value"">" & domainInfo("opModeForest") & "</TD></TR>"
	report.WriteLine "<TR><TD>Domnen-Betriebsmodus</TD><TD class=""value"">" & domainInfo("opModeDomain") & "</TD></TR>"

	' Write schema versions
	report.WriteLine "<TR><TD>Schema-Version</TD><TD class=""value"">" & domainInfo("schemaVersion") & "</TD></TR>"
	If domainInfo("exchSchemaVersion") <> "" Then report.WriteLine "<TR><TD>Exchange-Schema-Version</TD><TD class=""value"">" & domainInfo("exchSchemaVersion") & "</TD></TR>"
	If domainInfo("lyncSchemaVersion") <> "" Then report.WriteLine "<TR><TD>Lync-Schema-Version</TD><TD class=""value"">" & domainInfo("lyncSchemaVersion") & "</TD></TR>"

	report.WriteLine "<TR><TD>NetBIOS-Domnenname</TD><TD class=""value"">" & domainInfo("netBiosName") & "</TD></TR>"
	report.WriteLine "<TR><TD>Domnen-SID</TD><TD class=""value"">" & domainInfo("sID") & "</TD></TR>"
	report.WriteLine "<TR><TD>Standard-Container fr Benutzerkonten</TD><TD class=""value"">" & domainInfo("legacyUsersContainer") & "</TD></TR>"
	report.WriteLine "<TR><TD>Standard-Container fr Computerkonten</TD><TD class=""value"">" & domainInfo("legacyComputersContainer") & "</TD></TR>"
	If cfg("includeUPNSuffixes") And domainInfo.Exists("forestUPNSuffixes") Then 
		report.WriteLine "<TR><TD>UPN-Suffixe im Forest</TD><TD class=""value"">"
		If IsArray(domainInfo("forestUPNSuffixes")) Then
			For i = 0 To UBound(domainInfo("forestUPNSuffixes"))
				report.WriteLine domainInfo("forestUPNSuffixes")(i) & "<br />"
			Next
			report.WriteLine "</TD></TR>"
		Else
			report.WriteLine domainInfo("forestUPNSuffixes") & "</TD></TR>"
		End If
	End If
	

	' Write deleted object settings
	report.WriteLine "<TR><TD>Tombstone Lifetime</TD><TD class=""value"">" 
	If domainInfo("tombstoneLifetime") = 0  Then
		report.Write "60 Tage"
	Else	
		report.Write domainInfo("tombstoneLifetime") & " Tage"
	End If
	report.Write "</TD></TR>"
	report.WriteLine "<TR><TD>Deleted Objects Lifetime</TD><TD class=""value"">" 
	If domainInfo("msDS-DeletedObjectLifetime") = 0  Then
		report.Write "(nicht definiert; gleich Tombstone Lifetime)"
	Else	
		report.Write domainInfo("msDS-DeletedObjectLifetime") & " Tage"
	End If
	report.Write "</TD></TR>"
	
	' Write AD Recycle Bin
	report.WriteLine "<TR><TD>AD-Papierkorb</TD><TD class=""value"">"
	If domainInfo("HasRecycleBin") Then
		report.Write "aktiviert"
	Else
		report.Write "nicht aktiviert"
	End If
	report.Write "</TD></TR>"
	
	' Write Privileged Access Management
	report.WriteLine "<TR><TD>Privileged Access Management</TD><TD class=""value"">"
	If domainInfo("HasPAM") Then
		report.Write "aktiviert"
	Else
		report.Write "nicht aktiviert"
	End If
	report.Write "</TD></TR>"
	
	' Write all DCs
	report.WriteLine "<TR><TD>DCs der Domne <span class=""value"">(" & UBound(domainInfo("allDCs")) + 1 & ")</span></TD><TD class=""value"">"
	For i = 0 To UBound(domainInfo("allDCs"))
		report.WriteLine domainInfo("allDCs")(i) & "<br />"
	Next
	report.WriteLine "</TD></TR>"
	
	' Write account lockout settings.
	report.WriteLine "<TR><TD>Kontensperrdauer</TD><TD class=""value"">" 
	Select Case domainInfo("lockoutDuration") 
		Case -1 report.Write "(Wert nicht lesbar)"
		Case 0 report.Write "nur manuelles Entsperren"
		Case Else report.Write domainInfo("lockoutDuration") & " Minuten"
	End Select
	report.Write "</TD></TR>" & vbNewLine & "<TR><TD>Kontensperrung nach</TD><TD class=""value"">" & domainInfo("lockoutThreshold") & " falschen Kennwrtern"
	If domainInfo("lockoutThreshold") = 0 Then report.Write " (keine Sperrungen)"
	report.Write "</TD></TR>" & vbNewLine & "<TR><TD>Sperrungszhler gelscht nach</TD><TD class=""value"">" & domainInfo("lockoutObservationWindow") & " Minuten"
	report.Write "</TD></TR>" & vbNewLine & "<TR><TD>Minimale Kennwortlnge</TD><TD class=""value"">" & domainInfo("minPwdLength") & " Zeichen"
	report.Write "</TD></TR>" & vbNewLine & "<TR><TD>Maximales Kennwortalter</TD><TD class=""value"">" & domainInfo("maxPwdAge") & " Tage"
	report.Write "</TD></TR>" & vbNewLine & "<TR><TD>Minimales Kennwortalter</TD><TD class=""value"">" & domainInfo("minPwdAge") & " Tage"
	report.Write "</TD></TR>" & vbNewLine & "<TR><TD>Kennwortchronik</TD><TD class=""value"">" & domainInfo("pwdHistoryLength") & " alte Kennwrter"
	report.Write "</TD></TR>" & vbNewLine & "<TR><TD>Kennwort-Eigenschaften</TD><TD class=""value"">" & DecodePwdProperties(domainInfo("pwdProperties"))
	report.WriteLine "</TD></TR>" & vbNewLine & "</TABLE>"
	report.WriteLine "<div>&nbsp;</div>"
End Sub

' ------------------------------------------------------------------------------
' Write given trusts to report.
' ------------------------------------------------------------------------------
Sub WriteTrusts(report, arrTrusts)
	Dim i
	For i = LBound(arrTrusts) To UBound(arrTrusts)
		report.WriteLine "<div style=""margin-left:" & IndentionStep & "px;"">" _
			& symbols("Domain") & " Vertrauensstellung: <I>" & arrTrusts(i)("name") _
			& " (" & arrTrusts(i)("flatName") & ")" & vbNewLine _
			& "<UL style=""margin-left:" & IndentionStep & "px;""><LI>" & arrTrusts(i)("type") & "</LI>"
		If Not IsNull(arrTrusts(i)("direction")) Then report.WriteLine "<LI>" & arrTrusts(i)("direction") & "</LI>"
		If Not IsNull(arrTrusts(i)("attributes")) Then report.WriteLine "<LI>Attribute: " & arrTrusts(i)("attributes") & "</LI>"
		report.WriteLine "</UL></I></div>"
	Next
End Sub

' ------------------------------------------------------------------------------
' Write the given OU data to the report.
' ------------------------------------------------------------------------------
Sub WriteOUInfo(report, ou, allGPOs, lockoutDuration)
	Dim indent1, indent2, category, obj, child, i

	indent1 = ou("level") * IndentionStep
	indent2 = indent1 + IndentionStep

	PrintDebug "Write " & ou("name") & " data ..."

	' Don't write top-level element, because WriteDomainInfo() already did that.
	If ou("level") > 0 Then
		report.Write "<P style=""margin-left:" & indent1 & "px;"">" _
			& symbols(ou("class")) & "<span class=""OULevel"" title=""OU-Ebene " & ou("level") & """>" & ou("level") & "</span>&nbsp;<B>" & ou("name") & "</B>"
		WriteDescription report, ou("description")
		report.WriteLine "</P>"
		If cfg("includeUPNSuffixes") And ou.Exists("OUUPNSuffixes") Then 
			report.WriteLine "<P style=""margin-left:" & indent2 & "px;"">UPN-Suffixe der OU: <I>"
			If IsArray(ou("OUUPNSuffixes")) Then
				For i = 0 To UBound(ou("OUUPNSuffixes"))
					report.Write ou("OUUPNSuffixes")(i) & " "
				Next
			Else
				report.Write ou("OUUPNSuffixes")
			End If
			report.Write "</I></P>"
		End If
		If cfg("showParent") Then
			report.WriteLine "<P style=""margin-left:" & indent2 & "px;"">Eltern-Objekt: <I>" & ou("parent") & "</I></P>"
		End If
		If cfg("includeDates") Then
			report.WriteLine "<P style=""margin-left:" & indent2 & "px;"">Objekt erzeugt am: <I>" & ou("whenCreated") & "</I></P>"
			report.WriteLine "<P style=""margin-left:" & indent2 & "px;"">Objekt gendert am: <I>" & ou("whenChanged") & "</I></P>"
		End If
	End If

	If cfg("includeCounts") And ou.Exists("count") And ou("level") > 0 Then WriteCount report, ou("count"), indent2
	If cfg("includeGPOs") And ou.Exists("GPO") Then WriteLinkedGPOInfo report, ou("GPO"), ou("gpoInherit"), allGPOs, indent2
	If cfg("includeUsers") And ou.Exists("User") Then WriteObjectInfo report, "User", ou, lockoutDuration, indent2
	If cfg("includeComputers") And ou.Exists("Computer") Then WriteObjectInfo report, "Computer", ou, Null, indent2
	If cfg("includeGroups") And ou.Exists("Group") Then WriteObjectInfo report, "Group", ou, Null, indent2
	If cfg("includeContacts") And ou.Exists("Contact") Then WriteObjectInfo report, "Contact", ou, Null, indent2
	If cfg("includePrinters") And ou.Exists("Printer") Then WriteObjectInfo report, "Printer", ou, Null, indent2

	For Each child In ou("children")
		WriteOUInfo report, child, allGPOs, lockoutDuration
	Next
End Sub

' ------------------------------------------------------------------------------
' Write object counts to the report.
' ------------------------------------------------------------------------------
Sub WriteCount(report, count, indent)
	Dim category

	If IsEmpty(count) Then Exit Sub
	If count Is Nothing Then Exit Sub
	' should indent be 20 it's the domain level where we don't want object counts
	'If intent = 20 Then Exit Sub

	PrintDebug vbTab & "Write counts ..."

	report.WriteLine "<P style=""margin-left:" & indent _
		& "px;"">Anzahl Objekte: " & count("total") & "<UL>"
	If count("total") > 0 Then
		For Each category In count.Keys
			If category <> "total" And count(category) > 0 Then
				report.WriteLine "<LI style=""margin-left:" & indent + IndentionStep _
					& "px;"">Anzahl Objekte vom Typ """ & category & """: " _
					& count(category) & "</LI>"
			End If
		Next
	End If
	report.WriteLine "</UL></P>"
End Sub

' ------------------------------------------------------------------------------
' Write GPOs that are linked to the current container to the report.
' ------------------------------------------------------------------------------
Sub WriteLinkedGPOInfo(report, linkedGPOs, inherit, allGPOs, indent)
	Dim gpo
	Dim intLinkOrder

	If UBound(linkedGPOs) < 0 Then Exit Sub
	PrintDebug "Write linked group policies ..."

	If Not inherit Then report.WriteLine "<P style=""margin-left:" & indent _
		& "px;"">" & symbols("blocked") _
		& "Vererbung von Gruppenrichtlinien ist deaktiviert</P>"
	intLinkOrder = UBound(linkedGPOs) + 1
	For Each gpo In linkedGPOs
		PrintDebug vbTab & "Linked policy: " & gpo("guid")
		If allGPOs.Exists(gpo("guid")) Then
			' report only if GPO has been collected before
			report.Write "<P style=""margin-left:" & indent & "px;""><A href=""GPMC-Reports\" & allGPOs(gpo("guid"))("displayName") & ".html"">" & symbols("GPO") & "</A>"
			If gpo("disabled") Then report.Write symbols("disabled")
			If gpo("enforced") Then report.Write symbols("Exclamation")
			report.WriteLine "<span class=""GPOPrio"" title=""Prioritt " & intLinkOrder & """>Prio " & intLinkOrder & "</span> <B>" & allGPOs(gpo("guid"))("displayName") _
				& "</B>&nbsp;<I>(" & gpo("guid") & ")</I></P>"
		Else
			' GPO has not been collected - most likely a permission issue
			report.Write "<P style=""margin-left:" & indent & "px;"">" & symbols("GPO") _
				& "<span class=""GPOPrio"" title=""Prioritt " & intLinkOrder & """>Prio " & intLinkOrder & "</span> " _
				& symbols("Question") & "<B><I>(Zugriff nicht mglich)</B>&nbsp;(" & gpo("guid") & ")</I></P>"
			showPermissionHint = True
			SetPermissionReason "GPO-Zugriff " & gpo("guid")
		End If
		intLinkOrder = intLinkOrder - 1
	Next
End Sub

' ------------------------------------------------------------------------------
' Write the data of the given parent object collection to the report.
' ------------------------------------------------------------------------------
Sub WriteObjectInfo(report, category, parentObj, lockoutDuration, indent)
	Dim obj, unlockTime

	If parentObj.Exists("class") Then
		PrintDebug "Write " & LCase(category) & "s of " & parentObj("class") & " " & parentObj("name") & " ..."
	Else
		PrintDebug "Write " & LCase(category) & "s of host " & parentObj("name") & " ..."
	End If

	For Each obj In parentObj(category)
		PrintDebug vbTab & category & ": " & obj("name")
		report.Write "<P style=""margin-left:" & indent & "px;"">" & symbols(category)
		If category = "User" Or category = "Computer" Then
			If IsNull(obj("disabled")) Then
				report.Write symbols("Question")
			ElseIf obj("disabled") Then
				report.Write symbols("disabled")
			End If
		End If
		report.Write "<B>" & obj("name") & "</B>"
		If category = "User" Then
			If Not IsNull(obj("locked")) And obj("locked") And Not IsEmpty(lockoutDuration) Then
				report.Write "&nbsp;" & symbols("locked") & "gesperrt seit: " & obj("lockoutTime")
				If IsNull(obj("lockoutTime")) Then
					report.Write "(Wert nicht lesbar)"
				' lockoutDuration contains -1 when the value is not retrievable
				' in that case we should not calculate the unlock time
				ElseIf lockoutDuration > -1 Then	
					unlockTime = DateAdd("n", lockoutDuration, obj("lockoutTime"))
					If (Now > unlockTime) Then
						report.Write ", entsperrt seit: "
					Else
						report.Write ", wird entsperrt: "
					End If
					report.Write unlockTime
				Else
					report.Write " (Entsperrung nicht lesbar)"
				End If
			End If
			If obj("expires") <> #1/1/1601# Then report.Write "&nbsp;" & symbols("Clock") & "Konto luft ab: " & obj("expires") & " "
		End If
		WriteDescription report, obj("description")
		report.WriteLine "</P>"
		WriteAttributes report, obj("attributes"), (indent + IndentionStep)
		If obj.Exists("Printer") And cfg("includePrinters") Then WriteObjectInfo report, "Printer", obj, Null, (indent + IndentionStep)
	Next
End Sub

' ------------------------------------------------------------------------------
' Write the given descriptive comment to the report.
' ------------------------------------------------------------------------------
Sub WriteDescription(report, descr)
	If cfg("showDescriptions") And descr <> "" Then report.Write "&nbsp;<I>(" & descr & ")</I>"
End Sub

' ------------------------------------------------------------------------------
' Write the given attributes to the report.
' ------------------------------------------------------------------------------
Sub WriteAttributes(report, attributes, indent)
	Dim name, value, isNotDN, NumberOfValues

	' do not write a table when no attributes to report on
	If attributes.Count = 2 and not cfg("includeDates") Then Exit Sub

	report.WriteLine "<TABLE style=""margin-left:" & indent & "px;"">"
	For Each name In attributes.Keys
		' Write dates only if includeDates flag is set.
		If Not (LCase(name) = "whencreated" Or LCase(name) = "whenchanged") Or cfg("includeDates") Then
			isNotDN = (name <> "distinguishedName")
			VALUE = attributes(name)
			If cfg("showNaturalNames") Then name = Translate(name)
			If IsArray(value) Then
				Dim i
				' Get number of values (e.g. group members)
				NumberOfValues = UBound(value) + 1
				If NumberOfValues > 1 Then name = name & " (" & NumberOfValues & ")"
				' Walk through values
				For i = LBound(value) To UBound(value)
					WriteSingleAttribute report, name, VALUE(i), isNotDN
					' Don't write attribute name in subsequent lines.
					If i = LBound(value) Then name = "&nbsp;"
				Next
			Else
				WriteSingleAttribute report, name, value, isNotDN
			End If
		End If
	Next
	report.WriteLine "</TABLE>"
End Sub

' ------------------------------------------------------------------------------
' Write a given attribute to the report.
' ------------------------------------------------------------------------------
Sub WriteSingleAttribute(report, ByVal name, ByVal value, isNotDN)
	PrintDebug vbTab & vbTab & name & ": " & VALUE
	If VALUE <> "" Then
		If cfg("showFriendlyNames") And isNotDN And (InStr(1, value, "cn=", vbTextCompare) = 1) Then
			VALUE = Mid(value, 4, InStr(4, value, "=", vbTextCompare) - 7)
		End If
		report.WriteLine "<TR><TD width=""175"">" & name _
			& "</TD><TD class=""value"">" & VALUE & "</TD></TR>"
	End If
End Sub

' ------------------------------------------------------------------------------
' Write the given GPO data to the report.
' ------------------------------------------------------------------------------
Sub WriteGPOInfo(report, gpoList)
	Dim guid, link

	If gpoList.Count = 0 Then Exit Sub

	PrintDebug "Write group policy data ..."

	report.WriteLine "<H1>Liste der Gruppenrichtlinien (GPOs)</H1>"

	For Each guid In gpoList.Keys
		PrintDebug vbTab & "Policy: " & guid
		report.WriteLine "<P class=""policy""><A href=""GPMC-Reports\" & gpoList(guid)("displayName") & ".html"">" & symbols("GPO") & "</A><strong>" _
			& gpoList(guid)("displayName") & "</strong><BR>" & vbNewLine _
			& "<em>{" & guid & "}</em><BR>"

		If cfg("showDescriptions") And gpoList(guid)("comment") <> "" Then report.WriteLine "Kommentar:<BR><em>" & vbNewLine & Replace(gpoList(guid)("comment"), vbNewLine, "<BR>") & "</em><BR>"
		If gpoList(guid)("gpcWMIFilter") <> "" Then report.WriteLine "WMI-Filter: " & gpoList(guid)("gpcWMIFilter") & "<BR>"

		If cfg("includeDates") Then
			report.WriteLine "angelegt am: " & gpoList(guid)("whenCreated") & "<BR>"
			report.WriteLine "gendert am: " & gpoList(guid)("whenChanged") & "<BR>"
		End If

		report.Write "Version: " & gpoList(guid)("versionNumber")
		' GPO version numbers are stored in a DWORD with the user settings version
		' number being stored in the high WORD and the computer settings version
		' number being stored in the low WORD.
		report.WriteLine "&nbsp;(Computer: " & (gpoList(guid)("versionNumber") Mod 65536) _
			& ", User: " & (gpoList(guid)("versionNumber") \ 65536) & ")<BR>"

		report.Write "Status des Objekts: "
		Select Case gpoList(guid)("flags")
			' "flags" is a bitmap:
			' bit #0 = 0: user settings are enabled
			' bit #0 = 1: user settings are disabled
			' bit #1 = 0: computer settings are enabled
			' bit #1 = 1: computer settings are disabled
			Case 0 report.Write "Aktiviert"
			Case 1 report.Write "Benutzereinstellungen deaktiviert"
			Case 2 report.Write "Computereinstellungen deaktiviert"
			Case 3 report.Write "Alle Einstellungen deaktiviert"
			Case Else report.Write "unbekannt"
		End Select

		If cfg("includeCounts") Then report.Write "<BR>" & vbNewLine _
			& "Links in der dokumentierten Struktur: " _
			& (UBound(gpoList(guid)("linksDomain")) _
			+ UBound(gpoList(guid)("linksSite")) + 2)
		If UBound(gpoList(guid)("linksDomain")) >= 0 Then
			report.WriteLine "<BR>" & vbNewLine _
				& "<em>Links in der Domne:</em>" & vbNewLine _
				& "<UL class=""policy"">"
			For Each link In gpoList(guid)("linksDomain")
				report.WriteLine "<LI>" & link & "</LI>"
			Next
			report.WriteLine "</UL>"
		End If
		If UBound(gpoList(guid)("linksSite")) >= 0 Then
			report.WriteLine "<BR>" & vbNewLine _
				& "<em>Links an AD-Standorten:</em>" & vbNewLine _
				& "<UL class=""policy"">"
			For Each link In gpoList(guid)("linksSite")
				report.WriteLine "<LI>" & link & "</LI>"
			Next
			report.WriteLine "</UL>"
		End If

		report.WriteLine "</P>"
	Next

	If cfg("includeCounts") Then report.WriteLine "<P>Gruppenrichtlinienobjekte insgesamt: " & gpoList.Count & "</P>"
End Sub

' ------------------------------------------------------------------------------
' Write the given PSO data to the report.
' ------------------------------------------------------------------------------
Sub WritePSOs(report, PSOArray)
	Dim i
	Dim PSOAppliesTo
	
	PrintDebug "Getting PSOs ..."
	' check if we have any data
	' there could be none if no PSO is present or if permissions are insufficient
	If IsEmpty(PSOArray) then Exit Sub
	For i = 0 To UBound(PSOArray)
		PrintDebug vbTab & "PSO: " & PSOArray(i)("name")
		report.WriteLine "<P class=""pso"">" & symbols("PSO") & "<strong>" _
			& PSOArray(i)("name") & "</strong><BR>"
		report.WriteLine "<TABLE style=""margin-left:20px;"">"
		report.WriteLine "<TR><TD>Minimale Kennwortlnge</TD><TD><em> " & PSOArray(i)("msDS-MinimumPasswordLength") & " Zeichen</em></TD></TR>"
		report.WriteLine "<TR><TD>Maximales Kennwortalter</TD><TD><em> " & PSOArray(i)("msDS-MaximumPasswordAge") & " Tage</em></TD></TR>"
		report.WriteLine "<TR><TD>Kennwortchronik</TD><TD><em> " & PSOArray(i)("msDS-PasswordHistoryLength") & " alte Kennwrter</em></TD></TR>"
		report.WriteLine "<TR><TD>Kennwortkomplexitt</TD><TD><em> " & PSOArray(i)("msDS-PasswordComplexityEnabled") & "</em></TD></TR>"
		report.WriteLine "<TR><TD>Umkehrbare Verschlsselung</TD><TD><em> " & PSOArray(i)("msDS-PasswordReversibleEncryptionEnabled") & "</em></TD></TR>"
		report.WriteLine "<TR><TD>Kontensperrdauer</TD><TD><em> " & PSOArray(i)("msDS-lockoutDuration") & " Minuten</em></TD></TR>"
		report.WriteLine "<TR><TD>Kontensperrung nach</TD><TD><em> " & PSOArray(i)("msDS-lockoutThreshold") & " falschen Kennwrtern</em></TD></TR>"
		report.WriteLine "<TR><TD>Sperrungszhler gelscht nach</TD><TD><em> " & PSOArray(i)("msDS-lockoutObservationWindow") & " Minuten</em></TD></TR>"
		report.WriteLine "<TR><TD>PSO-Prioritt</TD><TD><em> " & PSOArray(i)("msDS-PasswordSettingsPrecedence") & "</em></TD></TR></TABLE>"
		If IsArray(PSOArray(i)("msDS-PSOAppliesTo")) Then 
			PSOAppliesTo = Join(PSOArray(i)("msDS-PSOAppliesTo"), "</LI><LI>")
		Else
			PSOAppliesTo = PSOArray(i)("msDS-PSOAppliesTo")
		End If
		report.WriteLine "<div style=""margin-left:20px;""> PSO gilt fr" & vbNewLine & "<em><UL class=""pso""><LI> " & PSOAppliesTo & "</LI></UL></em></div>"
		report.WriteLine "</P>"
	Next

End Sub


' ------------------------------------------------------------------------------
' Write caption to the report.
' ------------------------------------------------------------------------------
Sub WriteCaption(report)
	Dim objectClass, i
	If showPermissionHint = True Then
		report.WriteLine "<div class=""alertbox"">" & vbNewLine & "<H1>Hinweis zu Berechtigungen</H1>"
		report.WriteLine "<P>Bei der Auswertung verschiedener Objekte oder Attribute kann es " _
			& "Berechtigungsprobleme geben. Diese fhren dazu, dass Informationen zu deaktivierten, gesperrten " _
			& "oder ablaufenden Konten - ggf. aber auch andere Informationen - nicht richtig dargestellt werden.</P>" _
			& "<P>Um verlssliche Informationen zu diesen Objekten und Attributen zu erhalten, fhren Sie Jos bzw. JoseExec " _
			& "mit einem Konto aus, das volle Leserechte auf die relevanten Objekte und Attribute hat (z.B. ein Domnen-Administratorkonto). " _
			& "Ab Windows Vista und Windows Server 2008 sollten Sie Jos bzw. JoseExec dazu von einer Eingabeaufforderung " _
			& "aus starten, die Sie ausdrcklich als Administrator gestartet haben.</P>" _
			& "<P>(Nheres zu den Besonderheiten ab Vista finden Sie in " _
			& "<A HREF=""http://www.faq-o-matic.net/2008/02/22/benutzerkontensteuerung-uac-richtig-einsetzen/"">diesem</A> " _
			& "Artikel.)</P>"
		If cfg("showAdvancedPermissionHint") Then
			report.WriteLine "<P>Mglicherweise fehlende Berechtigungen:</P><UL class=""permission"">"
			For i = 0 To UBound(PermissionReason)
				report.WriteLine "<LI>" & PermissionReason(i) & "</LI>"
			Next
			report.WriteLine "</UL>" 
		End If
		report.WriteLine vbNewLine & "</div>"
	End If

	If cfg("includeCounts") And cfg("includeOUs") Then
		report.WriteLine "<H1 id=""stats"">Statistik dieses Reports</H1>"
		report.WriteLine "<P>Hinweis: Diese Statistik bercksichtigt nur die Objekte im Bereich dieses Reports."
		If cfg("objectFilter") <> "" Then report.WriteLine "(Ebenen-Filter: <em>" & cfg("objectFilter") & "</em>)"
		report.WriteLine "<TABLE>" & vbNewLine & "<TR><TH width=""150"">Objektklasse</TD><TH class=""count"">Anzahl</TD></TR>"
		For Each objectClass In stats.Keys
			If Left(objectClass, 3) <> "CN=" Then
				report.WriteLine "<TR><TD>" & objectClass & "</TD><TD class=""count"">" _
					& CStr(stats(objectClass)) & "</TD></TR>"
			End If
		Next
		If maxOULevel > 1 Then
		report.WriteLine "<TR><TD>Maximale OU-Tiefe</TD><TD class=""count"">" & maxOULevel & "</TD></TR>"
	End If
	report.WriteLine "</UL></P>"

		report.WriteLine "</TABLE>"
	End If

	report.WriteLine "<H1 id=""caption"">Legende</H1>" & vbNewLine & "<TABLE>"
	report.WriteLine "<TR><TD>" & symbols("Domain") & "</TD><TD><P>Domne</P></TD></TR>"
	report.WriteLine "<TR><TD>" & symbols("PSO") & "</TD><TD><P>Password Settings Object</P></TD></TR>"
	report.WriteLine "<TR><TD>" & symbols("organizationalUnit") & "</TD><TD><P>Organisationseinheit (OU)</P></TD></TR>"
	report.WriteLine "<TR><TD>" & symbols("GPO") & "</TD><TD><P>GPO-Verknpfung (Gruppenrichtlinie)</P></TD></TR>"
	report.WriteLine "<TR><TD>" & symbols("blocked") & "</TD><TD><P>GPO-Vererbung blockiert</P></TD></TR>"
	report.WriteLine "<TR><TD>" & symbols("Exclamation") & "</TD><TD><P>Anwendung der Gruppenrichtlinie wird erzwungen</P></TD></TR>"
	report.WriteLine "<TR><TD><span class=""GPOPrio"">Prio 1</span>&nbsp;</TD><TD><P>Prioritt des GPO-Links (1 = hchste Prioritt)</P></TD></TR>"
	report.WriteLine "<TR><TD>" & symbols("container") & "</TD><TD><P>Ordner (Container)</P></TD></TR>"
	report.WriteLine "<TR><TD>" & symbols("User") & "</TD><TD><P>Benutzer</P></TD></TR>"
	report.WriteLine "<TR><TD>" & symbols("Computer") & "</TD><TD><P>Computer</P></TD></TR>"
	report.WriteLine "<TR><TD>" & symbols("Group") & "</TD><TD><P>Gruppe</P></TD></TR>"
	report.WriteLine "<TR><TD>" & symbols("Contact") & "</TD><TD><P>Kontakt</P></TD></TR>"
	report.WriteLine "<TR><TD>" & symbols("Printer") & "</TD><TD><P>Drucker</P></TD></TR>"
	report.WriteLine "<TR><TD>" & symbols("disabled") & "</TD><TD><P>Objekt ist deaktiviert (manuell)<BR>Bei Gruppenrichtlinien: Link ist deaktiviert</P></TD></TR>"
	report.WriteLine "<TR><TD>" & symbols("locked") & "</TD><TD><P>Objekt ist gesperrt (vom System)</P></TD></TR>"
	report.WriteLine "<TR><TD>" & symbols("Question") & "</TD><TD><P>Status des Objekts unklar (verm. Berechtigungsfehler)</P></TD></TR>"
	report.WriteLine "<TR><TD>" & symbols("Clock") & "</TD><TD><P>Konto luft ab</P></TD></TR>"
	report.WriteLine "</TABLE>"
End Sub

' ------------------------------------------------------------------------------
' Write advertisements to the report.
' ------------------------------------------------------------------------------
Sub WriteAdv(report)
	report.WriteLine "<HR>" & vbNewLine & "<P id=""adv"">" _
		& symbols("Jos") & " Dokumentiert mit <strong>" & cfg("generator") & "</strong> am " _
		& Date & " um " & Time & ".<BR>" & vbNewLine _
		& "Download unter <A HREF=""" & Homepage & """>" _
		& symbols("FOM") & "</A>.</P>"
End Sub

' ==============================================================================
' Data gathering functions
' ==============================================================================

' ------------------------------------------------------------------------------
' Collect domain information on the given domain.
' ------------------------------------------------------------------------------
Function GetDomainInfo(domain)
	Dim domainInfo, objDomain
	Dim objSchema, objPartitions, objConfiguration, objDirectoryService, objRidManager, objInfrastructure
	Dim DCtoQuery
	Dim objUsersContainer, strUsersContainer
	Dim objComputersContainer, strComputersContainer
	Dim objExchangeSchemaVersion, objLyncSchemaVersion, objNetBIOS, objDuration, objlockoutObservationWindow
	Dim lockoutThreshold, lockoutDuration, lockoutObservationWindow, hi, lo
	Dim objMaxPwdAge, objMinPwdAge, minPwdLength, pwdHistoryLength, objPwdProperties
	Dim HasOptFeatures, HasRecycleBin, HasPAM
	Dim UPNForest
	Dim i
	PrintDebug "Get information about domain " & domain & " ..."

	Set domainInfo = CreateObject("Scripting.Dictionary")
	Set objDomain = GetObject("LDAP://" & domain)
	Set objConfiguration = GetObject("LDAP://" & rootDSE.Get("configurationNamingContext"))
	Set objSchema = GetObject("LDAP://" & rootDSE.Get("schemaNamingContext"))
	Set objPartitions = GetObject("LDAP://CN=Partitions," & rootDSE.Get("configurationNamingContext"))
	Set objDirectoryService = GetObject("LDAP://CN=Directory Service,CN=Windows NT,CN=Services," & rootDSE.Get("configurationNamingContext"))
	Set objRidManager = GetObject("LDAP://CN=RID Manager$,CN=System," & domain)
	Set objInfrastructure = GetObject("LDAP://CN=Infrastructure," & domain)

	domainInfo.Add "DCtoQuery", rootDSE.dnsHostName
	domainInfo.Add "lastBackup", GetADBackupDate(objDomain)
	domainInfo.Add "fqdn", DN2FQDN(domain)
	domainInfo.Add "description", objDomain.Description
	domainInfo.Add "whenCreated", objDomain.Get("whenCreated")
	domainInfo.Add "whenChanged", objDomain.Get("whenChanged")
	domainInfo.Add "rootDomain", DN2FQDN(rootDSE.get("rootDomainNamingContext"))
	domainInfo.Add "sID", Raw2Str(objDomain.Get("objectSID"))
	domainInfo.Add "allDCs", GetAllDCs(domain)
	UPNForest = ""
	On Error Resume Next
		UPNForest = objPartitions.Get("uPNSuffixes")
	On Error Goto 0
	' make sure that UPN is only added to domainInfo if it exists
	If (IsArray(UPNForest)) Then
		domainInfo.Add "forestUPNSuffixes", UPNForest
	ElseIf UPNForest <> "" Then
		domainInfo.Add "forestUPNSuffixes", UPNForest
	End If

	' Determine FSMO role owners.
	domainInfo.Add "fsmoSchemaMaster", GetName(objSchema.Get("fSMORoleOwner"))
	domainInfo.Add "fsmoDnsMaster", GetName(objPartitions.Get("fSMORoleOwner"))
	domainInfo.Add "fsmoPdcEmulator", GetName(objDomain.Get("fSMORoleOwner"))
	domainInfo.Add "fsmoRidMaster", GetName(objRidManager.Get("fSMORoleOwner"))
	domainInfo.Add "fsmoInfraMaster", GetName(objInfrastructure.Get("fSMORoleOwner"))

	' Forest mode didn't exist prior to Windows Server 2003, thus checking for
	' errors.
	On Error Resume Next

	' Determine operation modes.
	domainInfo.Add "opModeForest", "Betriebsmodus unbekannt (evtl. Windows-2000-Umgebung)"
	Select Case objPartitions.Get("msDS-Behavior-Version")
		Case "0" domainInfo("opModeForest") = "Windows 2000 gemischt (DCs: NT, 2000, 2003)"
		Case "1" domainInfo("opModeForest") = "Windows Server 2003 interim (DCs: NT, 2003)"
		Case "2" domainInfo("opModeForest") = "Windows Server 2003 pur (DCs: 2003 oder neuer)"
		Case "3" domainInfo("opModeForest") = "Windows Server 2008 (DCs: 2008 oder neuer)"
		Case "4" domainInfo("opModeForest") = "Windows Server 2008 R2 (DCs: 2008 R2 oder neuer)"
		Case "5" domainInfo("opModeForest") = "Windows Server 2012 (DCs: 2012 oder neuer)"
		Case "6" domainInfo("opModeForest") = "Windows Server 2012 R2 (DCs: 2012 R2 oder neuer)"
		Case "7" domainInfo("opModeForest") = "Windows Server 2016 (DCs: 2016 oder neuer)"
		Case Else TrapError "GetDomainInfo() - forest operation mode"
	End Select
	' the attribute "msDS-Behavior-Version" does not exist in Windows 2000 - clear errors related to this
	if err.number <> 0 then Err.Clear

	domainInfo.Add "opModeDomain", "Betriebsmodus unbekannt (evtl. Windows-2000-Umgebung)"
	Select Case objDomain.Get("msDS-Behavior-Version") 
		Case "0" domainInfo("opModeDomain") = "Windows 2000"
		Case "1" domainInfo("opModeDomain") = "Windows Server 2003"
		Case "2" domainInfo("opModeDomain") = "Windows Server 2003"
		Case "3" domainInfo("opModeDomain") = "Windows Server 2008"
		Case "4" domainInfo("opModeDomain") = "Windows Server 2008 R2"
		Case "5" domainInfo("opModeDomain") = "Windows Server 2012"
		Case "6" domainInfo("opModeDomain") = "Windows Server 2012 R2"
		Case "7" domainInfo("opModeDomain") = "Windows Server 2016"
		Case Else TrapError "GetDomainInfo() - domain operation mode"
	End Select
	' the attribute "msDS-Behavior-Version" does not exist in Windows 2000 - clear errors related to this
	if err.number <> 0 then Err.Clear
	If objDomain.Get("ntMixedDomain") = "1" Then
		domainInfo("opModeDomain") = domainInfo("opModeDomain") & " (gemischt)"
	End If

	' Determine schema versions
	domainInfo.Add "schemaVersion", objSchema.Get("objectVersion")
	TrapError "GetDomainInfo() - AD schema version"
	Select Case domainInfo("schemaVersion")
		Case "13" domainInfo("schemaVersion") = domainInfo("schemaVersion") & " (Windows 2000 Server)"
		Case "30" domainInfo("schemaVersion") = domainInfo("schemaVersion") & " (Windows Server 2003)"
		Case "31" domainInfo("schemaVersion") = domainInfo("schemaVersion") & " (Windows Server 2003 R2)"
		Case "44" domainInfo("schemaVersion") = domainInfo("schemaVersion") & " (Windows Server 2008)"
		Case "46" domainInfo("schemaVersion") = domainInfo("schemaVersion") & " (Windows Server 2008 R2 Beta/RC)"
		Case "47" domainInfo("schemaVersion") = domainInfo("schemaVersion") & " (Windows Server 2008 R2)"
		Case "51" domainInfo("schemaVersion") = domainInfo("schemaVersion") & " (Windows Server 8 Developer Preview)"
		Case "52" domainInfo("schemaVersion") = domainInfo("schemaVersion") & " (Windows Server 8 Beta)"
		Case "56" domainInfo("schemaVersion") = domainInfo("schemaVersion") & " (Windows Server 2012)" ' has not changed from Release Candidate
		Case "69" domainInfo("schemaVersion") = domainInfo("schemaVersion") & " (Windows Server 2012 R2)"  ' has not changed from Preview
		Case "72" domainInfo("schemaVersion") = domainInfo("schemaVersion") & " (Windows Server Technical Preview)" 
		Case "81" domainInfo("schemaVersion") = domainInfo("schemaVersion") & " (Windows Server Technical Preview 2)"
		Case "82" domainInfo("schemaVersion") = domainInfo("schemaVersion") & " (Windows Server Technical Preview 3)"
		Case "85" domainInfo("schemaVersion") = domainInfo("schemaVersion") & " (Windows Server 2016 TP 4)"
		Case "87" domainInfo("schemaVersion") = domainInfo("schemaVersion") & " (Windows Server 2016)" ' has not changed from TP5
		Case "88" domainInfo("schemaVersion") = domainInfo("schemaVersion") & " (Windows Server 2019)" ' has not changed from Insider Preview, starting Build 17623
		Case Else domainInfo("schemaVersion") = domainInfo("schemaVersion") & " (unbekannte Version!)"
	End Select

	' Exchange schema version
	Set objExchangeSchemaVersion = GetObject("LDAP://CN=ms-Exch-Schema-Version-Pt," & objSchema.Get("distinguishedName"))
	If objExchangeSchemaVersion Is Nothing Then
		Err.Clear
	Else
		domainInfo.Add "exchSchemaVersion", objExchangeSchemaVersion.Get("rangeUpper")
		Dim exchangeSchemaVersion : exchangeSchemaVersion = objExchangeSchemaVersion.Get("rangeUpper")
		TrapError "GetDomainInfo() - Exchange schema version"
		If exchangeSchemaVersion <> "" Then
		' see http://social.technet.microsoft.com/wiki/contents/articles/2772.exchange-schema-versions-common-questions-answers.aspx
			Select Case domainInfo("exchSchemaVersion")
				Case "4397" domainInfo("exchSchemaVersion") = domainInfo("exchSchemaVersion") & " (Exchange 2000 Server)"
				Case "4406" domainInfo("exchSchemaVersion") = domainInfo("exchSchemaVersion") & " (Exchange 2000 Server SP3)"
				Case "6870" domainInfo("exchSchemaVersion") = domainInfo("exchSchemaVersion") & " (Exchange Server 2003)"
				Case "6936" domainInfo("exchSchemaVersion") = domainInfo("exchSchemaVersion") & " (Exchange Server 2003 SP3)"
				Case "10637" domainInfo("exchSchemaVersion") = domainInfo("exchSchemaVersion") & " (Exchange Server 2007 RTM)"
				Case "11116" domainInfo("exchSchemaVersion") = domainInfo("exchSchemaVersion") & " (Exchange Server 2007 SP1)"
				Case "14622" domainInfo("exchSchemaVersion") = domainInfo("exchSchemaVersion") & " (Exchange Server 2010 RTM oder Exchange Server 2007 SP2)"
				Case "14625" domainInfo("exchSchemaVersion") = domainInfo("exchSchemaVersion") & " (Exchange Server 2007 SP3)"
				Case "14726" domainInfo("exchSchemaVersion") = domainInfo("exchSchemaVersion") & " (Exchange Server 2010 SP1)"
				Case "14732" domainInfo("exchSchemaVersion") = domainInfo("exchSchemaVersion") & " (Exchange Server 2010 SP2)"
				Case "14734" domainInfo("exchSchemaVersion") = domainInfo("exchSchemaVersion") & " (Exchange Server 2010 SP3)"
				Case "15137" domainInfo("exchSchemaVersion") = domainInfo("exchSchemaVersion") & " (Exchange Server 2013 RTM)"
				Case "15254" domainInfo("exchSchemaVersion") = domainInfo("exchSchemaVersion") & " (Exchange Server 2013 CU1)"
				Case "15281" domainInfo("exchSchemaVersion") = domainInfo("exchSchemaVersion") & " (Exchange Server 2013 CU2)"
				Case "15283" domainInfo("exchSchemaVersion") = domainInfo("exchSchemaVersion") & " (Exchange Server 2013 CU3)"
				Case "15292" domainInfo("exchSchemaVersion") = domainInfo("exchSchemaVersion") & " (Exchange Server 2013 CU4)"
				Case "15300" domainInfo("exchSchemaVersion") = domainInfo("exchSchemaVersion") & " (Exchange Server 2013 CU5)"
				Case "15303" domainInfo("exchSchemaVersion") = domainInfo("exchSchemaVersion") & " (Exchange Server 2013 CU6)"
				Case "15312" domainInfo("exchSchemaVersion") = domainInfo("exchSchemaVersion") & " (Exchange Server 2013 CU7-14)"
				Case "15317" domainInfo("exchSchemaVersion") = domainInfo("exchSchemaVersion") & " (Exchange Server 2016 RTM)"
				Case "15323" domainInfo("exchSchemaVersion") = domainInfo("exchSchemaVersion") & " (Exchange Server 2016 CU1)"
				Case "15325" domainInfo("exchSchemaVersion") = domainInfo("exchSchemaVersion") & " (Exchange Server 2016 CU2)"
				Case "15326" domainInfo("exchSchemaVersion") = domainInfo("exchSchemaVersion") & " (Exchange Server 2016 CU3-5)"
				Case "15330" domainInfo("exchSchemaVersion") = domainInfo("exchSchemaVersion") & " (Exchange Server 2016 CU6)"
				Case "15332" domainInfo("exchSchemaVersion") = domainInfo("exchSchemaVersion") & " (Exchange Server 2016 CU7-12)"
				Case "17000" domainInfo("exchSchemaVersion") = domainInfo("exchSchemaVersion") & " (Exchange Server 2019 RTM oder CU1)"
				Case Else domainInfo("exchSchemaVersion") = domainInfo("exchSchemaVersion") & " (unbekannte Version!)"
			End Select
		End If
		Set objExchangeSchemaVersion = Nothing
	End If

	' Lync schema version
	' see http://www.newcotec.com/blog/entry/lync-schema-versionen
	Set objLyncSchemaVersion = GetObject("LDAP://CN=ms-RTC-SIP-SchemaVersion," & objSchema.Get("distinguishedName"))
	If objLyncSchemaVersion Is Nothing Then
		Err.Clear
	Else
		domainInfo.Add "lyncSchemaVersion", objLyncSchemaVersion.Get("rangeUpper")
		Dim lyncSchemaVersion : lyncSchemaVersion = objLyncSchemaVersion.Get("rangeUpper")
		TrapError "GetDomainInfo() - Lync schema version"
		If lyncSchemaVersion <> "" Then
			Select Case domainInfo("lyncSchemaVersion")
				Case "1006" domainInfo("lyncSchemaVersion") = domainInfo("lyncSchemaVersion") & " (Live Communication Server 2005)"
				Case "1007" domainInfo("lyncSchemaVersion") = domainInfo("lyncSchemaVersion") & " (Office Communication Server 2007)"
				Case "1008" domainInfo("lyncSchemaVersion") = domainInfo("lyncSchemaVersion") & " (Office Communication Server 2007 R2)"
				Case "1100" domainInfo("lyncSchemaVersion") = domainInfo("lyncSchemaVersion") & " (Lync Server 2010)"
				Case "1150" domainInfo("lyncSchemaVersion") = domainInfo("lyncSchemaVersion") & " (Lync Server 2013/Skype for Business 2015)"
				Case Else domainInfo("lyncSchemaVersion") = domainInfo("lyncSchemaVersion") & " (Lync/Skype for Business - unbekannte Version!)"
			End Select
		End If
		Set objLyncSchemaVersion = Nothing
	End If

	' NetBIOS name is read from the crossRef object in the partition container.
	Set objNetBIOS = QueryAD("SELECT nETBIOSName FROM '" & objPartitions.aDSPath _
		& "' " & "WHERE nCName = '" & domain & "'")
	domainInfo.Add "netBiosName", UCase(objNetBIOS("nETBIOSName"))
	TrapError "GetDomainInfo() - NetBIOS name"
	Set objNetBIOS = Nothing
	On Error Goto 0

	' Well-known containers are stored in a multi-value attribute in the domain head
	' they can be bound to using a special syntax, see http://msdn.microsoft.com/en-us/library/windows/desktop/ms675565%28v=vs.85%29.aspx
	' and http://support.microsoft.com/kb/324949/en-us?lc=1031
	' this seems not to have been there in Windows 2000, so avoid errors
	On Error Resume Next
	' first: the legacy users container
	Set objUsersContainer = GetObject("LDAP://<WKGUID=a9d1ca15768811d1aded00c04fd8d5cd," & domain & ">")
	strUsersContainer = objUsersContainer.Get("distinguishedName")
	Set objUsersContainer = Nothing
	domainInfo.Add "legacyUsersContainer", strUsersContainer
	PrintDebug "Legacy Users Container: " & strUsersContainer
	' second: the legacy computers container
	Set objComputersContainer = GetObject("LDAP://<WKGUID=AA312825768811D1ADED00C04FD8D5CD," & domain & ">")
	strComputersContainer = objComputersContainer.Get("distinguishedName")
	Set objComputersContainer = Nothing
	domainInfo.Add "legacyComputersContainer", strComputersContainer
	PrintDebug "Legacy Computers Container: " & strComputersContainer
	On Error Goto 0
	

	' *** Get settings for deleted objects
	' settings are empty by default, so suppress errors
	PrintDebug "Getting settings for deleted objects ..."
	On Error Resume Next
	domainInfo.Add "tombstoneLifetime", 0
	domainInfo("tombstoneLifetime") = objDirectoryService.Get("tombstoneLifetime")
	domainInfo.Add "msDS-DeletedObjectLifetime", 0
	domainInfo("msDS-DeletedObjectLifetime") = objDirectoryService.Get("msDS-DeletedObjectLifetime")
	On Error Goto 0

	' *** Query Optional Features
	PrintDebug "Getting AD Optional Settings ..."
	domainInfo.Add "HasRecycleBin", False
	domainInfo.Add "HasPAM", False
	On Error Resume Next
	' after more than six years we finally have to extend the code here with Windows Server 2016
	' so that it supports more than one Optional Feature. ;)
	HasOptFeatures = objPartitions.Get("msDS-EnabledFeature")
	On Error Goto 0

	If IsArray(HasOptFeatures) Then
		For i = 0 To UBound(HasOptFeatures)
			If LCase(Left(HasOptFeatures(i),10)) = "cn=recycle" Then domainInfo("HasRecycleBin") = True
			If LCase(Left(HasOptFeatures(i),13)) = "cn=privileged" Then domainInfo("HasPAM") = True
		Next
	Else
		If LCase(Left(HasOptFeatures,10)) = "cn=recycle" Then domainInfo("HasRecycleBin") = True
		If LCase(Left(HasOptFeatures,13)) = "cn=privileged" Then domainInfo("HasPAM") = True
	End If

	
	' *** Read Password Policy Settings ***
	' it seems safe to read these settings from the domain object as there apparently
	' is a process that takes the effective settings from the GPOs and writes them to
	' the domain head. Moreover, there is even code that writes changes to the domain head
	' back to the Default Domain Policy (no matter if the effective settings come from there ...)
	' compare: http://x220.minasi.com/forum/topic.asp?whichpage=1&TOPIC_ID=20647&#93786
	PrintDebug "Getting password settings ..."
	' Get account lockout settings.
	domainInfo.Add "lockoutThreshold", objDomain.Get("lockoutThreshold")

	domainInfo.Add "lockoutDuration", 0
	' there might be problems with lockoutDuration, so we'll catch errors
	' as we need the value later on
	' (well, at least I ran across an error once ...)
	Set objDuration = objDomain.lockoutDuration
	domainInfo("lockoutDuration") = Integer8Number(objDuration)
	If IsEmpty(domainInfo("lockoutDuration")) Then
	' make sure we don't get a Null value
		domainInfo("lockoutDuration") = -1
		PrintDebug "Error retrieving lockoutDuration; setting value to -1"
		showPermissionHint = True
		SetPermissionReason "Sperrdauer Domne" 
	End If

	' Get account lockout observation settings.
	domainInfo.Add "lockoutObservationWindow", 0
	Set objlockoutObservationWindow = objDomain.lockoutObservationWindow
	domainInfo("lockoutObservationWindow") = Integer8Number(objlockoutObservationWindow)

	' Get minimum password age
	domainInfo.Add "minPwdAge", 0
	Set objMinPwdAge = objDomain.minPwdAge
	domainInfo("minPwdAge") = Abs(Integer8Number(objMinPwdAge) / 60 / 24)
	
	' Get maximum password age
	domainInfo.Add "maxPwdAge", 0
	Set objMaxPwdAge = objDomain.maxPwdAge
	domainInfo("maxPwdAge") = Abs(Integer8Number(objMaxPwdAge) / 60 / 24)
	
	' Get password restrictions
	domainInfo.Add("minPwdLength"), objDomain.minPwdLength
	domainInfo.Add("pwdHistoryLength"), objDomain.pwdHistoryLength
	domainInfo.Add("pwdProperties"), objDomain.pwdProperties
	
	' ### Count data in the root container unless an object filter is specified
	' hm ... seems we do not need that here
	'PrintDebug "objectFilter is '" & cfg("objectFilter") & "'"
	'If cfg("objectFilter") = "" then GetCount(objDomain)
	
	Set objInfrastructure = Nothing
	Set objRidManager = Nothing
	Set objPartitions = Nothing
	Set objSchema = Nothing
	Set objDomain = Nothing

	Set GetDomainInfo = domainInfo
End Function

' ------------------------------------------------------------------------------
' Get all DCs in the given domain.
' See:
' - http://www.selfadsi.de/extended-ad/search-domain-controllers.htm
' ------------------------------------------------------------------------------
Function GetAllDCs(domain)
	Dim ldapFilter, objectList, dcDN, dcDNS
	Dim ado
	Dim arrDCs, i
	
	printDebug "Getting all DCs for domain " & domain & " ..."
	ldapFilter = "(&(objectCategory=computer)(userAccountControl:1.2.840.113556.1.4.803:=8192))"

	Set ado = CreateObject("ADODB.Connection")
	ado.Provider = "ADSDSOObject"
	ado.Open "ADSearch" 
	Set objectList = ado.Execute("<LDAP://" & domain & ">;" & ldapFilter & ";distinguishedName,dnsHostName;subtree")

	arrDCs = Array()
	i=0
	While Not objectList.EOF
		ReDim preserve arrDCs(i)
		dcDN = objectList.Fields("distinguishedName")
		dcDNS = objectList.Fields("dnsHostName")
		arrDCs(i) = dcDNS & " (" & dcDN & ")"
		PrintDebug "... " & arrDCs(i)
		objectList.MoveNext
		i = i + 1
	Wend 
	GetAllDCs = arrDCs
End Function

' ------------------------------------------------------------------------------
' Get trusts for the given domain.
' Trust data structures are documented here:
' - http://msdn2.microsoft.com/en-us/library/aa392340.aspx
' - http://msdn2.microsoft.com/en-us/library/ms722477.aspx
' ------------------------------------------------------------------------------
Function GetTrusts(domain)
	Dim attributes, wql, objTrust, arrTrusts, i, TrustAttr, TrustAttrString

	PrintDebug "Get trusts for domain " & domain & " ..."

	GetTrusts = Array()

	attributes = "name,flatName,trustAttributes,trustDirection,trustType"
	wql = "SELECT " & attributes & " FROM 'LDAP://CN=System," & domain _
		& "' WHERE objectClass='trustedDomain'"
	On Error Resume Next
	Set objTrust = QueryAD(wql)
	If TrapError("GetTrusts(" & wql & ")") Then 
		showPermissionHint = True
		SetPermissionReason "Vertrauensstellungen"
		Exit Function
	End If
	On Error Goto 0

	If Not objTrust.EOF Then objTrust.MoveFirst
	arrTrusts = Array()
	i = 0
	Do Until objTrust.EOF
		ReDim Preserve arrTrusts(i)
		Set arrTrusts(i) = CreateObject("Scripting.Dictionary")
		arrTrusts(i).Add "name", objTrust("name").value
		PrintDebug "   Adding trust " & objTrust("name")
		arrTrusts(i).Add "flatName", objTrust("flatName").value

		arrTrusts(i).Add "direction", Null
		Select Case objTrust("trustDirection")
			Case 1 arrTrusts(i)("direction") = "eingehende Vertrauensbeziehung"
			Case 2 arrTrusts(i)("direction") = "ausgehende Vertrauensbeziehung"
			Case 3 arrTrusts(i)("direction") = "bidirektionale Vertrauensbeziehung"
		End Select

		arrTrusts(i).Add "type", "unbekannter Beziehungstyp"
		Select Case objTrust("trustType")
			Case 1 arrTrusts(i)("type") = "Beziehung zu einer Windows-NT-4.0-Domne"
			Case 2 arrTrusts(i)("type") = "Beziehung zu einer Active-Directory-Domne (Windows 2000 oder hher)"
			Case 3 arrTrusts(i)("type") = "Beziehung zu einem Kerberos-Realm"
			Case 4 arrTrusts(i)("type") = "Beziehung zu einem DCE-Realm"
		End Select
		
		arrTrusts(i).Add "attributes", Null
		TrustAttr = objTrust("trustAttributes").value
		arrTrusts(i)("attributes") = DecodeTrustFlags(TrustAttr) & " (" & TrustAttr & ")"

		i = i + 1
		objTrust.MoveNext
	Loop
	GetTrusts = arrTrusts
End Function

' ------------------------------------------------------------------------------
' Return a dictionary with the OUs of the given domain. The dictionary has the
' following structure:
'
' "class"       [string]     object class (can be either "builtinDomain",
'                            "container" or "organizationalUnit")
' "name"        [string]     the object's name
' "level"       [integer]    the object's level in the OU hierarchy
' "description" [string]     descriptive comment
' "whenCreated" [timestamp]  the object's creation time
' "whenChanged" [timestamp]  time when the object was last changed
' "count"       [dictionary] (optional) total and per-category counts of
'                            contained objects (users, computers, printers, ...)
' "gpoInherit"  [boolean]    are GPOs inherited from parent containers?
' "GPO"         [array]      GPOs assigned to the OU
' "User"        [array]      users assigned to the OU
' "Computer"    [array]      computers assigned to the OU
' "Group"       [array]      groups assigned to the OU
' "Contact"     [array]      contacts assigned to the OU
' "Printer"     [array]      printers assigned to the OU
' "children"    [array]      (optional) array of dictionaries of child objects
'                            of the same class
' ------------------------------------------------------------------------------
Function GetOUs(ouClass, dn, level)
	Dim objStart, objSub, thisOU, options

	PrintDebug ouClass & ": " & dn
	
	If level > maxOULevel then maxOULevel = level
	
	' catch error in case of missing permissions
	On Error Resume Next
	Set objStart = GetObject("LDAP://" & EscapeDN(dn))
	If TrapError("Error binding to object!") Then
		showPermissionHint = True
		SetPermissionReason "OU-Zugriff " & dn
		Exit Function
	End If
	On Error Goto 0

	Set thisOU = CreateObject("Scripting.Dictionary")

	thisOU.Add "class", ouClass
	thisOU.Add "name", objStart.Get("name")
	' if cfg ...
	thisOU.Add "parent", Replace(objStart.Parent, "LDAP://", "")
	thisOU.Add "level", level
	thisOU.Add "description", ""
	' Ignore error in case "description" is empty.
	On Error Resume Next
	thisOU("description") = objStart.Get("description")
	On Error Goto 0
	' Ignore error in case "uPNSuffixes" is empty.
	On Error Resume Next
	thisOU("OUUPNSuffixes") = objStart.Get("uPNSuffixes")
	On Error Goto 0
	thisOU.Add "whenCreated", objStart.Get("whenCreated")
	thisOU.Add "whenChanged", objStart.Get("whenChanged")
	thisOU.Add "gpoInherit", True
	On Error Resume Next
	options = &h00000000
	options = objStart.Get("gpOptions")
	thisOU("gpoInherit") = Not (CInt(options) And GPO_IGNORE)
	On Error Goto 0

	If InStr(1, dn, cfg("objectFilter"), vbTextCompare) Then
		If ouClass = "organizationalUnit" Or level > 0 Then
			' Get data of non-container objects if requested
			If cfg("includeCounts") Then thisOU.Add "count", GetCount(objStart)
			If cfg("includeGPOs") Then thisOU.Add "GPO", GetLinkedGPOs(objStart)
			If cfg("includeUsers") Then thisOU.Add "User", GetObjects("user", objStart, arrUserAttr)
			If cfg("includeComputers") Then thisOU.Add "Computer", GetObjects("computer", objStart, arrComputerAttr)
			If cfg("includeGroups") Then thisOU.Add "Group", GetObjects("group", objStart, arrGroupAttr)
			If cfg("includeContacts") Then thisOU.Add "Contact", GetObjects("contact", objStart, arrContactAttr)
			If cfg("includePrinters") Then thisOU.Add "Printer", GetObjects("printer", objStart, arrPrinterAttr)
		End If
	End If

	' Recurse only for objects of the same ouClass.
	objStart.Filter = Array(ouClass)
	If IsMatching(dn, cfg("objectFilter")) Then
		Dim i : i = 0
		Dim children : children = Array()
		For Each objSub In objStart
			If IsMatching(objSub.Get("distinguishedName"), cfg("objectFilter")) Then
				Dim name : name = objSub.Get("name")
				' Skip system containers, foreign security principals, Exchange
				' system objects and program data containers.
				Select Case LCase(name)
					Case "system"
					Case "foreignsecurityprincipals"
					Case "microsoft exchange system objects"
					Case "program data"
					Case Else
						ReDim Preserve children(i)
						'Set children(i) = GetOUs(ouClass, Replace(objSub.Get("distinguishedName"), "/", "\/"), level + 1)
						Set children(i) = GetOUs(ouClass, objSub.Get("distinguishedName"), level + 1)
						i = i + 1
				End Select
			End If
		Next
		thisOU.Add "children", children
	End If

	Set objStart = Nothing

	Set GetOUs = thisOU
End Function

' ------------------------------------------------------------------------------
' Return a dictionary with the number of objects in the given container (total
' as well as by category).
' ------------------------------------------------------------------------------
Function GetCount(objStart)
	Dim count, obj, objectClass

	Set GetCount = Nothing

	' this line prevented the function from counting objects in root
	' seems we should delete it
	'###If InStr(Join(objStart.Get("objectClass")), "domain") > 0 Then Exit Function

	' PrintDebug "Get counts for '" & objStart.Get("name") & "' ..."

	Set GetCount = CreateObject("Scripting.Dictionary")

	count = 0
	For Each obj In objStart
		count = count + 1
	Next
	GetCount.Add "total", count

	For Each objectClass In Array("user", "computer", "group", "contact", "printQueue", "organizationalUnit")
		objStart.Filter = Array(objectClass)
		count = 0
		For Each obj In objStart
			' Increment count unless a user actually is a computer.
			If Not (objectClass = "user" And IsComputer(obj.Get("objectCategory"))) Then count = count + 1
		Next
		GetCount.Add objectClass, count
		' Count objects for (global) statistics.
		stats(objectClass) = stats(objectClass) + count
		' PrintDebug "Count of " & objectClass & ": " & stats(objectClass)
	Next
End Function

' ------------------------------------------------------------------------------
' Return an array of dictionaries representing the GPOs linked to the given
' object. Each dictionary contains the following information:
'
' "guid"        [string]  the GPO's GUID
' "disabled"    [boolean] GPO is enabled/disabled for the object
' "enforced"    [boolean] GPO is or isn't enforced for the object
' ------------------------------------------------------------------------------
Function GetLinkedGPOs(objStart)
	Dim gpLinks, re, i, link, linkStatus

	Dim arrGPOs : arrGPOs = Array()

	gpLinks = ""
	On Error Resume Next
	gpLinks = Trim(objStart.Get("gPLink"))
	On Error Goto 0

	If gpLinks <> "" Then
		Set re = New RegExp
		re.Pattern = ".*?{(.+?)}.*?;(.+)"
		re.Global = True
		re.IgnoreCase = True

		If Left(gpLinks, 1) = "[" Then gpLinks = Mid(gpLinks, 2)
		If Right(gpLinks, 1) = "]" Then gpLinks = Mid(gpLinks, 1, InStrRev(gpLinks, "]") - 1)
		i = 0
		For Each link In Split(gpLinks, "][")
			PrintDebug vbTab & "Linked GPO: " & link
			If re.Test(link) Then
				ReDim Preserve arrGPOs(i)
				Set arrGPOs(i) = CreateObject("Scripting.Dictionary")

				Dim matches : Set matches = re.Execute(link)
				arrGPOs(i).Add "guid", matches(0).SubMatches(0)
				arrGPOs(i).Add "disabled", False
				arrGPOs(i).Add "enforced", False

				' "Enforce" bit MUST be ignored when "ignore" bit is set.
				' See: <http://msdn.microsoft.com/en-us/library/cc232505.aspx>
				linkStatus = CInt(matches(0).SubMatches(1))
				If linkStatus And GPO_IGNORE Then
					arrGPOs(i)("disabled") = True
				ElseIf linkStatus And GPO_ENFORCE Then
					arrGPOs(i)("enforced") = True
				End If
				Set matches = Nothing

				i = i + 1
			End If
		Next
	End If

	GetLinkedGPOs = arrGPOs
End Function

' ------------------------------------------------------------------------------
' Return an array of dictionaries representing the objects of the given
' category that are contained by objStart. Each dictionary contains the
' following elements:
'
' "name"        [string]     the object's name
' "description" [string]     descriptive comment
' "disabled"    [boolean]    (optional) is the object disabled?
' "expires"     [timestamp]  the object's expiry date (only category "user")
' "locked"      [boolean]    is the object locked?
' "lockoutTime" [timestamp]  (optional) the time the object was locked
' "attributes"  [dictionary] selected attributes of the object
' "Printer"     [array]      (only category "computer") print queues shared by
'                            the computer
' ------------------------------------------------------------------------------
Function GetObjects(category, objStart, attributes)
	Dim objSub, i

	PrintDebug "Get " & category & " objects in '" & objStart.Get("name") & "' ..."

	Dim arrObjects : arrObjects = Array()

	If category = "printer" Then
		objStart.Filter = Array("printQueue")
	Else
		objStart.Filter = Array(category)
	End If

	i = 0
	For Each objSub In objStart
		If category <> "user" Or Not IsComputer(objSub.Get("objectCategory")) Then
			PrintDebug vbTab & "Add " & objSub.Get("name")
			ReDim Preserve arrObjects(i)
			Set arrObjects(i) = CreateObject("Scripting.Dictionary")
			arrObjects(i).Add "name", objSub.Get("name")
			arrObjects(i).Add "description", ""
			On Error Resume Next
			arrObjects(i)("description") = objSub.Get("description")
			On Error Goto 0
			If category = "user" Or category = "computer" Then 
				arrObjects(i).Add "disabled", IsDisabled(objSub)
				arrObjects(i).Add "locked", IsLocked(objSub)
			End If	
			If category = "user" Then arrObjects(i).Add "expires", GetExpiryDate(objSub)
			If arrObjects(i)("locked") Then arrObjects(i).Add "lockoutTime", GetLockoutTime(objSub)
			arrObjects(i).Add "attributes", GetAttributes(objSub, attributes)
			If category = "computer" Then arrObjects(i).Add "Printer", GetObjects("printer", objSub, arrPrinterAttr)
			i = i + 1
		End If
	Next

	GetObjects = arrObjects
End Function

' ------------------------------------------------------------------------------
' Return an array with the specified attributes of the given object.
' ------------------------------------------------------------------------------
Function GetAttributes(objSub, attributes)
	Dim i, name, attribs

	PrintDebug vbTab & "Get attributes of " & objSub.Get("name") & " ..."

	Set attribs = CreateObject("Scripting.Dictionary")

	For i = LBound(attributes) To UBound(attributes)
		name = Trim(attributes(i))
		attribs.Add name, ""

		On Error Resume Next
		attribs(name) = objSub.Get(name)
		If Err.Number <> 0 Then
			' In case the attribute needs to be requested first.
			objSub.GetInfoEx Array(name), 0
			attribs(name) = objSub.Get(name)
		End If
		On Error Goto 0

		Select Case LCase(name)
			Case "" attribs.Remove(name)
			Case "objectsid"
				attribs(name) = Raw2Str(objSub.ObjectSid)
			Case "sidhistory"
				' sIDHistory is stored as an array.
				If IsArray(attribs(name)) Then
					If LCase(TypeName(attribs(name))) = "variant()" Then
						Dim j
						For j = LBound(attribs(name)) To UBound(attribs(name))
							attribs(name)(j) = Raw2Str(attribs(name)(j))
						Next
					Else
						attribs(name) = Raw2Str(attribs(name))
					End If
				End If
			Case "streetaddress"
				Dim postcode, city
				On Error Resume Next
				postcode = objSub.Get("postalCode")
				city = objSub.Get("l")
				On Error Goto 0
				If attribs(name) <> "" And (postcode <> "" Or city <> "") Then
					attribs(name) = Trim(attribs(name) & ", " & Trim(postcode & " " & city))
				Else
					attribs(name) = Trim(attribs(name) & " " & Trim(postcode & " " & city))
				End If
			Case "grouptype"
				Dim group
				If attribs(name) < 0 Then
					group = " (Sicherheitsgruppe)"
					attribs(name) = attribs(name) + 2147483648
				Else
					group = " (Verteilergruppe)"
				End If
				Select Case attribs(name)
					Case 2 attribs(name) = "Globale Gruppe"
					Case 4 attribs(name) = "Domnenlokale Gruppe"
					Case 5 attribs(name) = "Vordefinierte Gruppe"
					Case 8 attribs(name) = "Universalgruppe"
					Case Else attribs(name) = "Unbekannt (" & attribs(name) & ")"
				End Select
				attribs(name) = attribs(name) & group
			Case "dcroles"
				attribs(name) = GetDCRoles(objSub)
			Case "lastlogontimestamp"
				' this exists only in forest mode 2003 or above
				' so we need to handle errors
				On Error Resume Next
				attribs(name) = Integer8Date(objSub.Get(name), systemBias)
				If Err.number <> 0 Then
					attribs(name) = ""
					Err.Clear
				End If
				On Error Goto 0
			Case "logonhours"
				attribs(name) = GetLogonHours(objSub)
			Case "pwdlastset"
				On Error Resume Next
				attribs(name) = Integer8Date(objSub.Get(name), systemBias)
				if attribs(name) = #1/1/1601# then attribs(name) = "niemals"
				On Error Goto 0
			Case "useraccountcontrol"
				Dim flags : flags = 0
				On Error Resume Next
				flags = objSub.UserAccountControl
				On Error Goto 0
				If flags <> 0 Then
					attribs(name) = CStr(flags) & ": " & DecodeUACFlags(flags)
				Else
					attribs(name) = ""
				End If
			Case "terminalserviceshomedirectory"
				' this exists only in Windows Server 2003 and later
				On Error Resume Next
				attribs(name) = objSub.TerminalServicesHomeDirectory
				On Error Goto 0
			Case "terminalservicesprofilepath"
				' this exists only in Windows Server 2003 and later
				On Error Resume Next
				attribs(name) = objSub.TerminalServicesProfilePath
				On Error Goto 0
			Case "altrecipient"
				' differentiate between mail forward and forward-and-copy
				On Error Resume Next
				If objSub.deliverAndRedirect Then
					attribs("altrecipientcopy") = objSub.altRecipient
					attribs(name) = ""
				Else
					attribs(name) = objSub.altRecipient
				End If
				On Error Goto 0
			Case "msexchgenericforwardingaddress"
				' differentiate between mail forward and forward-and-copy
				On Error Resume Next
				If objSub.deliverAndRedirect Then
					attribs("msexchgenericforwardingaddresscopy") = objSub.msExchGenericForwardingAddress
					attribs(name) = ""
				Else
					attribs(name) = objSub.msExchGenericForwardingAddress
				End If
				On Error Goto 0
		End Select
		If IsArray(attribs(name)) Then
			PrintDebug vbTab & vbTab & name & " = " & Join(attribs(name), ",")
		Else
			PrintDebug vbTab & vbTab & name & " = " & attribs(name)
		End If
	Next

	Set GetAttributes = attribs
End Function

' ------------------------------------------------------------------------------
' Return a dictionary of dictionaries. The parent dictionary contains child
' dictionaries representing the GPOs defined in the given domain, which are
' identified by the GPO's respective GUID. Each child dictionary contains the
' following elements:
'
' "displayName"   [string]    the name the GPO is displayed as in the GUI
' "whenCreated"   [timestamp] the GPO's creation time
' "whenChanged"   [timestamp] time when the GPO was last changed
' "versionNumber" [DWORD]     the GPO's version number
' "flags"         [integer]   bitmap defining if computer/user settings are
'                             enabled or disabled
' "comment"       [string]    descriptive comment
' "gpcWMIFilter"  [string]    (optional) name of the WMI filter associated with
'                             the GPO
' "linksDomain"   [array]     locations where the GPO is linked in the domain
' "linksSite"     [array]     locations where the GPO is linked in the site
'
' This is a dictionary of dictionaries instead of an array of dictionaries,
' because when writing the information which GPO is linked to a particular
' container, properties of the respective GPO are read from this collection.
' Accessing a dictionary for that matter is far more efficient than accessing
' an array.
' ------------------------------------------------------------------------------
Function GetGPOs(domain)
	Dim attributes, query, objGPO, theGPOs, guid

	PrintDebug "Get group policy objects ..."

	attributes = "aDSPath,displayName,name,whenCreated,whenChanged,versionNumber" _
		& ",flags,gPCFileSysPath"
	query = "SELECT " & attributes & " FROM 'LDAP://CN=Policies,CN=System," & domain _
		& "' WHERE objectClass = 'groupPolicyContainer' ORDER BY displayName"

	On Error Resume Next
	Set objGPO = QueryAD(query)
	If TrapError("GetGPOs()") Or objGPO Is Nothing Then Exit Function

	Set theGPOs = CreateObject("Scripting.Dictionary")

	Do Until objGPO.EOF
		guid = Replace(Replace(objGPO("name").Value, "{", ""), "}", "")
		PrintDebug vbTab & "GPO: " & guid
		theGPOs.Add guid, CreateObject("Scripting.Dictionary")

		theGPOs(guid).Add "displayName", objGPO("displayName").Value
		theGPOs(guid).Add "whenCreated", objGPO("whenCreated").Value
		theGPOs(guid).Add "whenChanged", objGPO("whenChanged").Value
		theGPOs(guid).Add "versionNumber", objGPO("versionNumber").Value
		theGPOs(guid).Add "flags", objGPO("flags").Value
		theGPOs(guid).Add "comment", GetGPOComment(objGPO("gPCFileSysPath").Value)
		theGPOs(guid).Add "gpcWMIFilter", GetWMIFilter(objGPO("adsPath").Value)

		If Err.Number <> 0 Then Err.Clear

		theGPOs(guid).Add "linksDomain", GetGPOLinks(objGPO("name").Value, domain)
		theGPOs(guid).Add "linksSite", GetGPOLinks(objGPO("name").Value, rootDSE.Get("configurationNamingContext"))

		objGPO.MoveNext
	Loop
	On Error Goto 0

	Set GetGPOs = theGPOs
End Function

' ------------------------------------------------------------------------------
' Return data on the last AD backup: date, time, and performing DC
' This data is stored in two attributes, one of which is XML-formatted - 
' we'll read this attribute
' ------------------------------------------------------------------------------
Function GetADBackupDate(objDom)
	Dim ReplValue, Subvalue, DSASignature
	Dim timeLeft, timeRight, DCLeft, DCRight
	Dim TimeVal, DCName, BackupDate, BackupTime
	
	' the attribute was introduced in Windows 2003, so we'll catch errors
	On Error Resume Next
	objDom.GetInfoEx Array("msDS-ReplAttributeMetaData"),0
	ReplValue = objDom.Get("msDS-ReplAttributeMetaData")
	If Err.number <> 0 Then 
		GetADBackupDate = "(keine Information vorhanden)"
		Exit Function
	End If
	On Error Goto 0

	DSASignature = ""
	For Each Subvalue In ReplValue
		if instr(LCase(SubValue), "dsasignature") >0 then DSASignature = Subvalue
	Next
	' this is not too pretty but after all we do not need an XML parser here
	If DSASignature <> "" Then
		timeLeft = InStr(LCase(DSASignature), LCase("<ftimeLastOriginatingChange>")) + Len("<ftimeLastOriginatingChange>")
		timeRight = InStr(timeLeft, LCase(DSASignature), LCase("</ftimeLastOriginatingChange>"))
		DCLeft = InStr(LCase(DSASignature), LCase("<pszLastOriginatingDsaDN>")) + Len("<pszLastOriginatingDsaDN>") + Len("CN=NTDS Settings,CN=")
		DCRight = InStr(DCLeft, DSASignature, ",")
		TimeVal = Mid(DSASignature, timeLeft, timeRight - timeLeft)
		On Error Resume Next
		' DC Name might be missing
		DCName = Mid(DSASignature, DCLeft , DCRight - DCLeft)
		if err.number <> 0 then DCName = "(kein Wert)"
		On Error Goto 0
		BackupDate = Left(TimeVal, InStr(TimeVal, "T") - 1)
		BackupTime = Mid(TimeVal, InStr(TimeVal, "T") + 1)
		BackupTime = Left(BackupTime, Len(BackupTime) - 1)
		BackupTime = CDate(BackupDate & " " & BackupTime)
		BackupTime = DateAdd("n", -(systemBias), BackupTime)
		GetADBackupDate = BackupTime & " von " & DCName
	Else
		' AD seems not to have been backed up yet
		GetADBackupDate = "bislang nicht gesichert"
	End If
	
End Function

' ------------------------------------------------------------------------------
' Return an array of dictionaries with the Password Settings Objects defined In the domain.
' The array of dictionaries contains the following data:
' - single values for: name, msDS-MaximumPasswordAge, msDS-MinimumPasswordAge,
'   msDS-MinimumPasswordLength, msDS-PasswordHistoryLength, msDS-PasswordComplexityEnabled, 
'   msDS-PasswordReversibleEncryptionEnabled, msDS-lockoutObservationWindow, 
'   msDS-lockoutDuration, msDS-lockoutThreshold, msDS-PasswordSettingsPrecedence
' - multiple values (array) for: msDS-PSOAppliesTo
' ------------------------------------------------------------------------------
Function GetPSOs(domain)
	Dim PSOContainer, PSO, thePSOs, i
	
	' this requires admin rights, and of course the feature might be absent, so catch errors
	On Error Resume Next
	Set PSOContainer = GetObject("LDAP://CN=Password Settings Container,CN=System," & domain)
	If Err.number <> 0 Then
		showPermissionHint = True
		SetPermissionReason "Zugriff PSO"
		Exit Function
	End If
	On Error Goto 0
	
	PSOContainer.Filter = Array("msDS-PasswordSettings")
	thePSOs = Array()
	i = 0
	For Each PSO In PSOContainer
		ReDim Preserve thePSOs(i)
		Set thePSOs(i) = CreateObject("Scripting.Dictionary")
		thePSOs(i).Add "name", PSO.Get("name")
		thePSOs(i).Add "msDS-MinimumPasswordLength", PSO.Get("msDS-MinimumPasswordLength")
		thePSOs(i).Add "msDS-PasswordHistoryLength", PSO.Get("msDS-PasswordHistoryLength")
		thePSOs(i).Add "msDS-PasswordComplexityEnabled", PSO.Get("msDS-PasswordComplexityEnabled")
		thePSOs(i).Add "msDS-PasswordReversibleEncryptionEnabled", PSO.Get("msDS-PasswordReversibleEncryptionEnabled")
		thePSOs(i).Add "msDS-lockoutThreshold", PSO.Get("msDS-lockoutThreshold")
		thePSOs(i).Add "msDS-PasswordSettingsPrecedence", PSO.Get("msDS-PasswordSettingsPrecedence")
		thePSOs(i).Add "msDS-PSOAppliesTo", PSO.Get("msDS-PSOAppliesTo")
		thePSOs(i).Add "msDS-MaximumPasswordAge", Integer8Number(PSO.Get("msDS-MaximumPasswordAge"))/60/24
		thePSOs(i).Add "msDS-MinimumPasswordAge", Integer8Number(PSO.Get("msDS-MinimumPasswordAge"))/60/24
		thePSOs(i).Add "msDS-lockoutObservationWindow", Integer8Number(PSO.Get("msDS-lockoutObservationWindow"))
		thePSOs(i).Add "msDS-lockoutDuration", Integer8Number(PSO.Get("msDS-lockoutDuration"))
		i = i + 1
	Next
	GetPSOs = thePSOs
	Set PSOContainer = Nothing

End Function


' ==============================================================================
' Helper functions
' ==============================================================================

' ------------------------------------------------------------------------------
' Create a dictionary for storing attributes that should be read from AD.
' Include attributes "whenCreated" and "whenChanged" by default.
' ------------------------------------------------------------------------------
Function CreateAttributeSet()
	Set CreateAttributeSet = CreateObject("Scripting.Dictionary")
	CreateAttributeSet.Add "whenCreated", True
	CreateAttributeSet.Add "whenChanged", True
End Function

' ------------------------------------------------------------------------------
' Evaluate a definition line. If the definition has the form "key = value" an
' array [key, value] is returned. If the definition is a flag (i.e. just "key"
' without a "value") then an array [key, True] is returned.
' ------------------------------------------------------------------------------
Function EvalDefinition(line)
	Dim def : def = Split(line, "=", 2)
	def(KEY) = Trim(def(KEY))
	If UBound(def) = KEY Then
		ReDim Preserve def(VALUE)
		def(VALUE) = True
	Else
		def(VALUE) = Trim(def(VALUE))
	End If
	EvalDefinition = def
End Function

' ------------------------------------------------------------------------------
' Check whether an account is disabled or not. Return True if the account is
' disabled, False if it's disabled, Null if the attribute can't be read for
' some reason.
' ------------------------------------------------------------------------------
Function IsDisabled(objAccount)
	Dim status : status = 0

	On Error Resume Next
	Err.Clear
	status = objAccount.UserAccountControl
	If Err.Number <> 0 Or status = 0 Then
		' perhaps a permissions error
		showPermissionHint = True
		SetPermissionReason "Aktivierungsstatus " & objAccount.Name
		IsDisabled = Null
	ElseIf (status And ADS_UF_ACCOUNTDISABLE) Then
		IsDisabled = True
	Else
		IsDisabled = False
	End If
	On Error Goto 0
End Function

' ------------------------------------------------------------------------------
' Return the expiry date of the given object or Null if an error occurs.
' ------------------------------------------------------------------------------
Function GetExpiryDate(objSub)
	On Error Resume Next
	Err.Clear
	GetExpiryDate = Integer8Date(objSub.AccountExpires, systemBias)
	If Err.Number <> 0 Then
		PrintDebug "Can't get expiry date for object " & objSub.Get("name")
		' perhaps a permissions error
		showPermissionHint = True
		SetPermissionReason "Ablaufzeit " & objSub.Get("name")
		GetExpiryDate = Null
	End If
	On Error Goto 0
End Function

' ------------------------------------------------------------------------------
' Check whether an account is disabled or not. Return True if the account is
' disabled, False if it's disabled, Null if the attribute can't be read for
' some reason.
' ------------------------------------------------------------------------------
Function IsLocked(objAccount)
	Dim status

	On Error Resume Next
	Err.Clear
	status = objAccount.UserAccountControl
	If Err.Number <> 0 Or status = 0 Then
		' perhaps a permissions error
		showPermissionHint = True
		SetPermissionReason "Sperrstatus " & objAccount.Name
		IsLocked(0) = Null
	ElseIf (status And ADS_UF_LOCKOUT) Then
		IsLocked = True
	Else
		IsLocked = False
	End If
End Function

' ------------------------------------------------------------------------------
' Return the time the given account was locked out or Null in case of an error.
' ------------------------------------------------------------------------------
Function GetLockoutTime(objAccount)
	On Error Resume Next
	Err.Clear
	GetLockoutTime = Integer8Date(objAccount.LockoutTime, systemBias)
	If TrapError("GetLockoutTime()") Then 
		GetLockoutTime = Null
		showPermissionHint = True
		SetPermissionReason "Sperrzeit fr " & objAccount.Name
	End If
	On Error Goto 0
End Function

' ------------------------------------------------------------------------------
' Return the logon hours for the given account.
' Based on <http://www.rlmueller.net/Document%20LogonHours.htm>
' ------------------------------------------------------------------------------
Function GetLogonHours(objAccount)
	Dim DayNames
	Dim AllLogonHours(20)
	Dim LogonHoursBits(167)
	Dim LogonHours
	Dim LogonHourByte, LogonHour
	Dim k, Counter, LoopCounter, j, m
	Dim DayOfWeek(6), DayLogon(6)
	Dim DayCount, HourCount

	' German day names (short), beginning with Sunday (US-like)
	DayNames = Array("So", "Mo", "Di", "Mi", "Do", "Fr", "Sa")
	
	' Retrieve the user's logonHours attribute.
	On Error Resume Next
	objAccount.GetInfoEx Array("logonHours"), 0
	LogonHours = objAccount.Get("logonHours")
	' no logon hours for the account - return
	'   (this happens usually if the object was created via script -
	'   effectively, this means logon is allowed any time. However,
	'   we will not return any value)
	If Err.Number <> 0 Then Exit Function
	On Error Goto 0
	
	' Populate a byte array.
	For k = 1 To LenB(LogonHours)
		AllLogonHours(k - 1) = AscB(MidB(LogonHours, k, 1))
	Next
	
	' Populate a bit array, offset by the time zone bias.
	j = 0
	For Each LogonHourByte In AllLogonHours
		For k = 7 To 0 Step -1
			m = 8*j + k - (systemBias/60)
			If (m < 0) Then
				m = m + 168
			End If
			' *** correction Nils Kaczenski:
			' wrap the other way as well!
			If (m > 167) Then
				m = m - 168
			End If
			If (LogonHourByte And 2^k) Then
				LogonHoursBits(m) = 1
			Else
				LogonHoursBits(m) = 0
			End If
		Next
		j = j + 1
	Next
	
	Counter = 0
	LoopCounter = 0
	DayCount = 0
	
	For Each LogonHour In LogonHoursBits
		' add logon hours to an array of days
		DayOfWeek(DayCount) = DayOfWeek(DayCount) & LogonHour
		Counter  = Counter + 1
		If (Counter = 24) Then
			' this makes a full day, so check it
			If DayOfWeek(DayCount) = String(24, "1") Then 
				' User may log on the whole day
				DayLogon(DayCount) = "immer"
			Else
				' User has at least one exclusion, so walk through the hours
				' start counting at 1 so we can use the InStr 0 value
				' (but we need to adjust this in the code)
				HourCount = 1
				Do Until HourCount = 25
					' find times the user may not log on
					HourCount = InStr(HourCount, DayOfWeek(DayCount), "0")
					If HourCount = 0 Then 
						' no more exclusions, day complete
						HourCount = 25
					Else
						' add beginning of exclusion time
						DayLogon(DayCount) = DayLogon(DayCount) & ", nicht " & HourCount - 1
						' find next allowed hour
						HourCount = InStr(HourCount, DayOfWeek(DayCount), "1")
						If HourCount = 0 Then ' Or HourCount > 24
							' no more allowed hours, day complete
							HourCount = 25
							DayLogon(DayCount) = DayLogon(DayCount) & " bis 24"
						Else
							' add ending of exclusion time
							DayLogon(DayCount) = DayLogon(DayCount) & " bis " & HourCount - 1
							HourCount = HourCount + 1
						End If
					End If
				Loop
			End If	
			' Clear the beginning of the string
			If Left(DayLogon(DayCount), 2) = ", " Then DayLogon(DayCount) = Mid(DayLogon(DayCount), 3)
			DayCount = DayCount + 1
			Counter = 0
			LoopCounter = LoopCounter + 1
		End If
	Next
	If Join(DayOfWeek, "") = String(168, "1") Then
		' User may logon any time
		GetLogonHours = "jederzeit"
	ElseIf Join(DayOfWeek, "") = String(168, "0") Then
		' User may never log on
		GetLogonHours = "niemals"
	Else
		' Output Monday to Saturday
		For DayCount = 1 To 6
			GetLogonHours = GetLogonHours & DayNames(DayCount) & ": " & DayLogon(DayCount) & " <BR />" & vbNewLine
		Next
		' Outside the US, Sunday is the last day, not the first one
		GetLogonHours = GetLogonHours & DayNames(0) & ": " & DayLogon(0)
	End If
	
End Function

' ------------------------------------------------------------------------------
' Obtain local timezone bias from machine registry.
' ------------------------------------------------------------------------------
Function GetSystemBias
	Dim biasKey, k

	Dim objShell : Set objShell = CreateObject("WScript.Shell")
	biasKey = objShell.RegRead("HKLM\System\CurrentControlSet\Control\TimeZoneInformation\ActiveTimeBias")
	Set objShell = Nothing

	If (LCase(TypeName(biasKey)) = "long") Then
		GetSystemBias = biasKey
	ElseIf (LCase(TypeName(biasKey)) = "variant()") Then
		GetSystemBias = 0
		For k = 0 To UBound(biasKey)
			GetSystemBias = GetSystemBias + (biasKey(k) * 256^k)
		Next
	End If
End Function

' ------------------------------------------------------------------------------
' Collect reasons why permissions might not be granted
' ------------------------------------------------------------------------------
Function SetPermissionReason(ByVal Reason)
	If Not cfg("showAdvancedPermissionHint") Then Exit Function
	Dim i
	i = UBound(PermissionReason) + 1
	ReDim Preserve PermissionReason(i)
	PermissionReason(i) = Reason
End Function


' ------------------------------------------------------------------------------
' Escape illegal characters in a DN.
' It seems that the forward slash is the only character that is not auto-escaped.
' AD seems to accept a slash in an object name, however, using such a DN results
' in a bind error. That's why we need this function. Maybe it even has to be extended.
' ------------------------------------------------------------------------------
Function EscapeDN(ByVal DN)
	EscapeDN = Replace(DN, "/", "\/")
End Function

' ------------------------------------------------------------------------------
' Convert Integer8 (64-bit) value to a date, adjusted for local time zone bias.
' ------------------------------------------------------------------------------
Function Integer8Date(ByVal objDate, ByVal bias)
	Dim hi, lo
	hi = objDate.HighPart
	lo = objDate.LowPart
	' Account for bug in IADslargeInteger property methods.
	If (lo < 0) Then hi = hi + 1
	If (hi = 0) And (lo = 0) Then bias = 0
	On Error Resume Next
	Err.Clear
	Integer8Date = CDate(#1/1/1601# + (((hi * 2^32) + lo) / 600000000 - bias) / 1440)
	If Err.Number <> 0 Then Integer8Date = #1/1/1601#
	On Error Goto 0
End Function

' -------------------------------------------
' Convert Integer8 (64-bit) value to a number
' -------------------------------------------
Function Integer8Number(ByVal objInteger)
	Dim hi, lo
	hi = objInteger.HighPart
	lo = objInteger.LowPart
	Set objInteger = Nothing
	' Account for bug in IADslargeInteger property methods.
	If (lo < 0) Then hi = hi + 1
	If lo <> 0 Or hi <> -(2^31) Then
		Integer8Number = -(hi * 2^32 + lo) / 600000000
	End If
End Function

' ------------------------------------------------------------------------------
' Determine whether the given computer object is a domain controller or not.
' ------------------------------------------------------------------------------
Function GetDCRoles(objComputer)
	Dim options, objRID, objRS, objNtds

	' Only DCs have an RID set.
	' Attention: We might have to change this as it might not find RODCs!
	Set objRID = Nothing
	On Error Resume Next
	Set objRID = GetObject("LDAP://CN=RID Set," & objComputer.Get("distinguishedName"))
	On Error Goto 0

	If Not objRID Is Nothing Then
		' Computer is a domain controller.
		' Check if it's also a global catalog server.
		options = 0
		Set objRS = QueryAD("SELECT distinguishedName FROM 'LDAP://" _
			& rootDSE.Get("configurationNamingContext") & "' " _
			& "WHERE objectClass = 'server' and name='" & objComputer.Get("name") & "'")
		If Not objRS.EOF Then
			Set objNtds = GetObject("LDAP://CN=NTDS Settings," _
				& objRS("distinguishedName"))
			On Error Resume Next
			options = objNtds.Get("options")
			On Error Goto 0
		End If

		If options And NTDSDSA_OPT_IS_GC Then
			'
			GetDCRoles = Array("Domnencontroller", "Global Catalog Server")
		Else
			GetDCRoles = "Domnencontroller"
		End If
	Else
		' Computer is a domain member.
		GetDCRoles = "Domnenmitglied"
	End If
End Function

' ------------------------------------------------------------------------------
' Decode the TrustAttributes flags.
' See <http://msdn.microsoft.com/en-us/library/ms680832.aspx>
' ------------------------------------------------------------------------------
Function DecodeTrustFlags(trustAttr)
	Dim flag, arrFlags, i

	DecodeTrustFlags = ""
	If Not IsNumeric(trustAttr) Then Exit Function

	arrFlags = Array()
	i = 0
	For Each flag In trustFlags.Keys
		If TrustAttr And flag Then
			ReDim Preserve arrFlags(i)
			arrFlags(i) = trustFlags(flag)
			i = i + 1
		End If
	Next

	DecodeTrustFlags = Join(arrFlags, ", ")
End Function

' ------------------------------------------------------------------------------
' Decode the UserAccountControl flags.
' See <http://msdn.microsoft.com/en-us/library/ms680832.aspx>
' ------------------------------------------------------------------------------
Function DecodeUACFlags(uac)
	Dim flag, arrFlags, i

	DecodeUACFlags = ""
	If Not IsNumeric(uac) Then Exit Function

	arrFlags = Array()
	i = 0
	For Each flag In uacFlags.Keys
		If uac And flag Then
			ReDim Preserve arrFlags(i)
			arrFlags(i) = uacFlags(flag)
			i = i + 1
		End If
	Next

	DecodeUACFlags = Join(arrFlags, ", ")
End Function

' ------------------------------------------------------------------------------
' Decode the pwdProperties flags.
' See <http://msdn.microsoft.com/en-us/library/cc245476(PROT.13).aspx>
' ------------------------------------------------------------------------------
Function DecodePwdProperties(pwdProp)

	DecodePwdProperties = ""
	If Not IsNumeric(pwdProp) Then Exit Function

	If pwdProp And DOMAIN_PASSWORD_COMPLEX Then DecodePwdProperties = "komplexe Kennwrter"
	If pwdProp And DOMAIN_PASSWORD_NO_CLEAR_CHANGE Then DecodePwdProperties = DecodePwdProperties & ", keine Klartext-Kennwortnderungen"
	If pwdProp And DOMAIN_PASSWORD_STORE_CLEARTEXT Then DecodePwdProperties = DecodePwdProperties & ", Kennwrter entschlsselbar speichern"
	
	If Left(DecodePwdProperties,2) = ", " Then DecodePwdProperties = Mid(DecodePwdProperties, 2)
	
	if trim(DecodePwdProperties) = "" then DecodePwdProperties = "keine Vorgaben"
	
End Function

' ------------------------------------------------------------------------------
' Translate a given canonical name into a "natural" one.
' ------------------------------------------------------------------------------
Function Translate(name)
	Translate = name
	If naturalName.Exists(LCase(name)) Then Translate = naturalName(LCase(name))
End Function

' ------------------------------------------------------------------------------
' Read GPO comment from file.
' ------------------------------------------------------------------------------
Function GetGPOComment(path)
	GetGPOComment = ""
	If path <> "" Then
		path = fso.BuildPath(path, "GPO.cmt")
		If fso.FileExists(path) Then
			Dim f: Set f = fso.OpenTextFile(path, ForReading, False, TristateUseDefault)
			GetGPOComment = f.ReadAll
			f.Close
			Set f = Nothing
		End If
		TrapError "GetGPOComment()"
	End If
End Function

' ------------------------------------------------------------------------------
' Get WMI filter (if present).
' ------------------------------------------------------------------------------
Function GetWMIFilter(adsPath)
	PrintDebug vbTab & "Checking WMI Filter for " & adsPath
	Dim objGPOWMI : Set objGPOWMI = GetObject(adsPath)
	GetWMIFilter = ""
	GetWMIFilter = objGPOWMI.Get("gpcWQLFilter")
	If GetWMIFilter <> "" Then GetWMIFilter = WMIFilterName(GetWMIFilter)
End Function

' ------------------------------------------------------------------------------
' Get the WMI filter name matching the given GUID name
' ------------------------------------------------------------------------------
Function WMIFilterName(WMIGUID)
	PrintDebug vbTab & "Getting WMI Filter friendly name for " & WMIGUID
	Dim WMIGUIDPos, WMIGUIDEnd
	Dim WMICont, objWMI
	' Isolate GUID
	WMIGUIDPos = InStr(WMIGUID, "{")
	WMIGUIDEnd = InStr(WMIGUID, "}")
	WMIGUID = Mid(WMIGUID, WMIGUIDPos + 1, WMIGUIDEnd - WMIGUIDPos - 1)
    ' Browse through WMI objects
	Set WMICont = GetObject("LDAP://CN=SOM,CN=WMIPolicy,CN=System," & rootDSE.Get("defaultNamingContext"))
    WMICont.filter = Array("msWMI-Som")
    For Each objWMI In WMICont
		If InStr(objWMI.name, WMIGUID) > 0 Then
			' Found it, now get its name
			WMIFilterName = objWMI.get("msWMI-Name")
        End If
    Next
    Set WMICont = Nothing
	PrintDebug vbTab & "Found WMI Filter name " & WMIFilterName

End Function

' ------------------------------------------------------------------------------
' Get links of the named GPO in the given structure.
' ------------------------------------------------------------------------------
Function GetGPOLinks(name, structure)
	Dim fname, query, i, dn

	fname = "GetGPOLinks(" & name & "," & structure & ")"
	query = "SELECT distinguishedName FROM 'LDAP://" & structure _
		& "' WHERE gPLink='*" & name & "*' ORDER BY name"

	Dim arrGPOLinks : arrGPOLinks = Array()
	Dim objGPOLinks : Set objGPOLinks = QueryAD(query)
	TrapError fname & " - Get links from AD"

	i = 0
	Do Until objGPOLinks.EOF
		' Store DN of GPO link in array if the object has a readable DN.
		Err.Clear
		dn = objGPOLinks("distinguishedName").Value
		If Not TrapError(fname & " - Read DN") Then
			ReDim Preserve arrGPOLinks(i)
			arrGPOLinks(i) = dn
			i = i + 1
		End If
		objGPOLinks.MoveNext
	Loop

	Set objGPOLinks = Nothing
	GetGPOLinks = arrGPOLinks
End Function

' ------------------------------------------------------------------------------
' Run the given WQL query against AD using the globally established ADO
' connection.
' ------------------------------------------------------------------------------
Function QueryAD(wql)
	Set QueryAD = Nothing
	adoCmd.CommandText = wql
	Dim adoRS : Set adoRS = adoCmd.Execute
	If Not TrapError("QueryAD(" & wql & ")") Then Set QueryAD = adoRS
End Function

' ------------------------------------------------------------------------------
' Return True if either dn is a substring of filter or vice versa. Otherwise
' return False.
' ------------------------------------------------------------------------------
Function IsMatching(dn, Filter)
	If (InStr(1, dn, Filter, vbTextCompare) + InStr(1, Filter, dn, vbTextCompare)) > 0 Then
		IsMatching = True
	Else
		IsMatching = False
	End If
End Function

' ------------------------------------------------------------------------------
' Return True if the given category is "computer", otherwise return False.
' ------------------------------------------------------------------------------
Function IsComputer(category)
	IsComputer = (LCase(Left(category, 38)) = "cn=computer,cn=schema,cn=configuration")
End Function

' ------------------------------------------------------------------------------
' Check if cleartext ADSI error messages are available.
' ------------------------------------------------------------------------------
Function CheckForADSIErrorMessages()
	CheckForADSIErrorMessages = False
	Dim ADError : Set ADError = Nothing
	On Error Resume Next
	Set ADError = CreateObject("ADsError")
	On Error Goto 0
	If Not ADError Is Nothing Then CheckForADSIErrorMessages = True
End Function

' ------------------------------------------------------------------------------
' Format a domain's distinguished name as an FQDN (dot-notation).
' ------------------------------------------------------------------------------
Function DN2FQDN(dn)
	DN2FQDN = Replace(Replace(dn, ",DC=", "."), "DC=", "")
End Function

' ------------------------------------------------------------------------------
' Get the name of the given FSMO role owner.
' ------------------------------------------------------------------------------
Function GetName(fSMORoleOwner)
	Dim objNTDS : Set objNtds = GetObject("LDAP://" & fSMORoleOwner)
	Dim objComputer : Set objComputer = GetObject(objNtds.Parent)
	GetName = Replace(objComputer.Name, "CN=", "")
	Set objNtds = Nothing
	Set objComputer = Nothing
End Function

' ------------------------------------------------------------------------------
' Convert the given octet array into a string representation of its integer
' equivalent.
' ------------------------------------------------------------------------------
Function Octet2Str(arr)
	Dim i

	Octet2Str = 0
	For i = LBound(arr) To UBound(arr)
		Octet2Str = Octet2Str * 256 + arr(i)
	Next

	Octet2Str = CStr(Octet2Str)
End Function

' ------------------------------------------------------------------------------
' Convert an SID from (raw) binary string format into human-readable "S-..."
' notation.
'
' The binary encoding of an SID S-a-b-c-d-e-f-... is aNbbbbbbccccddddeeeeffff...
' where the respective octets are:
'
' a       (1 octet, revision number)
' N       (1 octet, number of DWORDs following the first 8 octets)
' bbbbbb  (6 octets, 48-bit number "b" in big-endian order)
' cccc    (DWORD (4 octets), 32-bit number "c" in little-endian order)
' dddd    (DWORD (4 octets), 32-bit number "d" in little-endian order)
' eeee    (DWORD (4 octets), 32-bit number "e" in little-endian order)
' ffff    (DWORD (4 octets), 32-bit number "f" in little-endian order)
' ...
'
' See: <http://blogs.msdn.com/oldnewthing/archive/2004/03/15/89753.aspx>
'
' For a list of well-known SIDs see <http://support.microsoft.com/kb/243330>.
' ------------------------------------------------------------------------------
Function Raw2Str(rawSID)
	Dim octets, i

	' Convert binary string to octet array.
	ReDim octets(LenB(rawSid))
	For i = 1 To LenB(rawSid)
		octets(i-1) = AscB(MidB(rawSid, i, 1))
	Next

	' Convert octet array to (human-readable) SID string.
	Raw2Str = "S-" & CStr(octets(0)) & "-" & Octet2Str(Array(octets(2), octets(3), octets(4), octets(5), octets(6), octets(7)))
	For i = 8 To (4 * octets(1) + 4) Step 4
		Raw2Str = Raw2Str & "-" & Octet2Str(Array(octets(i+3), octets(i+2), octets(i+1), octets(i)))
	Next
End Function

' ------------------------------------------------------------------------------
' If an error has occurred: try to decode the error number, report both error
' number and error message, and return True. Otherwise return False.
' ------------------------------------------------------------------------------
Function TrapError(msg)
	TrapError = False
	If Err.number <> 0 Then
		If hasADSIErrorMessages Then
			Dim objError : Set objError = CreateObject("ADsError")
			msg = msg & ": [ADSI] " & objError.GetErrorMessage(Err.Number) _
				& " (0x" & Hex(Err.Number) & ")"
			Set objError = Nothing
		Else
			msg = msg & ": " & Err.Description & " (" & Err.Number & ")"
		End If
		WScript.StdErr.WriteLine msg
		PrintDebug msg
		Err.Clear
		TrapError = True
	End If
End Function

' ------------------------------------------------------------------------------
' Print the syntax of how the script has to be invoked.
' ------------------------------------------------------------------------------
Sub PrintSyntax()
	WScript.StdOut.WriteLine vbNewLine & cfg("generator")
	WScript.StdOut.WriteLine vbNewLine & "Syntax:" & vbNewLine _
		& "cscript.exe //u " & WScript.ScriptName & " [/d:<Definition>] [/r:<Report>] [/t:<Titel>]" & vbNewLine _
		& "  [/f:<Filter>] [/debug]" & vbNewLine & vbNewLine _
		& "  /d:    (optional) Name einer Jos-Definitionsdatei" & vbNewLine _
		& "         Wird dieser Parameter nicht angegeben, liest das Skript die" & vbNewLine _
		& "         Definitionen von der Standardeingabe." & vbNewLine _
		& "  /r:    (optional) Name der zu erzeugenden Report-Datei (HTML)" & vbNewLine _
		& "         Achtung: Die Datei wird ohne Rckfrage berschrieben!" & vbNewLine _
		& "  /t:    (optional) Titel des Berichts" & vbNewLine _
		& "  /f:    (optional) LDAP-Objektfilter (Container/OU)" & vbNewLine _
		& "  /debug (optional) Schaltet den Debug-Modus ein." & vbNewLine & vbNewLine _
		& "Bei Dateinamen nur den Dateinamen (ohne Pfad) angeben!"
	WScript.StdOut.WriteLine vbNewLine & "Beispiel:" & vbNewLine _
		& "cscript.exe //u " & WScript.ScriptName & " /d:""DomainData.txt"" /r:""MyDomainData.htm"" /t:""Meine Domne"" /f:""OU=EDV,OU=Benutzer,DC=domain,DC=tld"""
	WScript.Quit 0
End Sub

' ------------------------------------------------------------------------------
' Print the given message to STDOUT if debugging is enabled.
' ------------------------------------------------------------------------------
Sub PrintDebug(msg)
	If cfg("debugEnabled") Then WScript.StdOut.WriteLine "[" & Time() & "] " & msg
End Sub
